"""Bedrock weatherer"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np

from .._base import _BaseMover
from ...utils import convert_to_array


################################################################################
# Component


class BedrockWeatherer(_BaseMover):
    """Bedrock weathering in a StackedLayers.

    Reference
    ---------
    Tucker, G. E., & Slingerland, R. L. (1994)
        Erosional dynamics, flexural isostasy, and long‐lived escarpments: A numerical modeling study
        https://doi.org/10.1029/94JB00320
    Granjeon, D. (1996)
        Modélisation stratigraphique déterministe: Conception et applications d'un modèle diffusif 3D multilithologique
        https://theses.hal.science/tel-00648827
    """

    _name = "BedrockWeatherer"

    _unit_agnostic = True

    _info = {
        "bathymetric__depth": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The water depth under the sea",
        },
    }

    def __init__(
        self,
        grid,
        max_weathering_rate=1e-5,
        weathering_decay_depth=1.0,
        wave_base=20.0,
        porosity=0.0,
        bedrock_composition=1.0,
        fields_to_track=None,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        max_weathering_rate : float (m/time)
            The maximum weathering rate of the bedrock.
        weathering_decay_depth : float (m)
            The characteristic weathering depth.
        wave_base : float (m)
            The wave base, below which weathering decreases exponentially.
        porosity : float or array-like (-)
            The porosity of the weathered material after formation for one or
            multiple lithologies.
        bedrock_composition : float or array-like (-)
            The composition of the material is added to the StackedLayers from the
            bedrock.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(grid, fields_to_track)

        # Parameters
        self.max_weathering_rate = max_weathering_rate
        self.weathering_decay_depth = weathering_decay_depth
        self.wave_base = wave_base
        self._porosity = convert_to_array(porosity)
        self.bedrock_composition = convert_to_array(bedrock_composition)

        # Physical fields
        self._bathymetry = grid.at_node["bathymetric__depth"]

        # Fields for weathering
        self._weathering_depth = np.zeros((grid.number_of_nodes, 1))

    def _calculate_weathering_depth(self, dt):
        """
        Calculates the weathering depth over the continental and marine domains.
        """
        self._weathering_depth[:, 0] = self.max_weathering_rate * dt
        self._weathering_depth[self._bathymetry > 0.0, 0] *= np.exp(
            -self._bathymetry[self._bathymetry > 0.0] / self.wave_base
        )
        self._weathering_depth[self._grid.core_nodes, 0] *= np.exp(
            -self._stratigraphy.thickness / self.weathering_decay_depth
        )

    def run_one_step(self, dt, update_compatible=False, update=False):
        """Run the weatherer for one timestep, dt.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        update_compatible : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer at the bottom of the stack
            only if the new layer is compatible with the existing layer.
        update : bool
            If false, a new layer is addded at each iteration to the StackedLayers
            of the grid; otherwise, the first layer is simply updated to save
            memory.
        """
        core_nodes = self._grid.core_nodes

        self._calculate_weathering_depth(dt)
        self._sediment_thickness[core_nodes] = (
            self.bedrock_composition
            * self._weathering_depth[core_nodes]
            / (1.0 - self._porosity)
        )

        self._update_stratigraphy(dt, update_compatible, update, True)
        self._topography[core_nodes] += np.sum(
            self._porosity * self._sediment_thickness[core_nodes], axis=1
        )



"""Sea level calculator"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from functools import partial
import numpy as np
from scipy.interpolate import CubicSpline, PchipInterpolator, make_interp_spline
from landlab import Component


################################################################################
# Component


class SeaLevelCalculator(Component):
    """Calculate sea level based on interpolating from control points."""

    _name = "SeaLevelCalculator"

    _unit_agnostic = True

    _info = {
        "topographic__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "sea_level__elevation": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "grid",
            "doc": "Sea level elevation",
        },
        "bathymetric__depth": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The water depth under the sea",
        },
    }

    def __init__(
        self,
        grid,
        control_time,
        control_sea_level,
        interpolation="linear",
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        control_time : array
            The time at each control point.
        control_sea_level : array
            The sea level at each control point.
        interpolation : str
            The interpolation method, which can be:
                - 'linear' for linear interpolation (the default).
                - 'quadratic' for spline interpolation of second order.
                - 'cubic' for spline interpolation of third order.
                - 'monotonic' for monotonic spline interpolation of third order.
        """
        super().__init__(grid)

        self._grid.at_grid["sea_level__elevation"] = 0.
        self.initialize_output_fields()

        # Parameters
        self._time = 0.0

        # Physical fields
        self._topography = grid.at_node["topographic__elevation"]
        self._bathymetry = grid.at_node["bathymetric__depth"]

        # Interpolator
        if interpolation == "linear":
            self._interpolator = partial(
                np.interp, xp=control_time, fp=control_sea_level
            )
        elif interpolation == "quadratic":
            self._interpolator = make_interp_spline(
                control_time, control_sea_level, k=2
            )
        elif interpolation == "cubic":
            self._interpolator = CubicSpline(control_time, control_sea_level)
        elif interpolation == "monotonic":
            self._interpolator = PchipInterpolator(control_time, control_sea_level)

        # Initialize the fields
        self.run_one_step(0.0)

    def run_one_step(self, dt):
        """Run the calculator for one timestep, dt.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        """
        self._time += dt

        sea_level = self._interpolator(self._time)
        self._grid.at_grid["sea_level__elevation"] = sea_level

        np.subtract(
            sea_level,
            self._topography,
            out=self._bathymetry,
            where=(self._topography < sea_level),
        )



"""Water-driven router"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np

from .._base import _BaseRouter, _BaseStreamPower
from .cfuncs import calculate_sediment_influx


################################################################################
# Component


class WaterDrivenRouter(_BaseRouter, _BaseStreamPower):
    """Water-driven diffusion of a Landlab field in continental and marine domains
    based on a routing scheme.

    References
    ----------
    Granjeon, D. (1996)
        Modélisation stratigraphique déterministe: Conception et applications d'un modèle diffusif 3D multilithologique
        https://theses.hal.science/tel-00648827
    Shobe, C. M., Tucker, G. E., & Barnhart, K. R. (2017)
        The SPACE 1.0 model: A Landlab component for 2-D calculation of sediment transport, bedrock erosion, and landscape evolution
        https://doi.org/10.5194/gmd-10-4577-2017
    """

    _name = "WaterDrivenRouter"

    _info = {
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__receiver_proportions": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of proportion of flow sent to each receiver.",
        },
        "water__unit_flux_in": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m/s",
            "mapping": "node",
            "doc": "External volume water per area per time input to each node (e.g., rainfall rate)",
        },
        "sediment__unit_flux_in": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m3/s",
            "mapping": "node",
            "doc": "Sediment flux as boundary condition",
        },
        "surface_water__discharge": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m3/s",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
        "bathymetric__depth": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The water depth under the sea",
        },
    }

    def __init__(
        self,
        grid,
        transportability_cont=1e-5,
        transportability_mar=1e-6,
        wave_base=20.0,
        critical_flux=0.0,
        porosity=0.0,
        max_erosion_rate_sed=1e-2,
        active_layer_rate_sed=None,
        bedrock_composition=1.0,
        max_erosion_rate_br=1e-2,
        active_layer_rate_br=None,
        exponent_discharge=1.0,
        exponent_slope=1.0,
        ref_water_flux=None,
        fields_to_track=None,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        transportability_cont : float or array-like (m/time)
            The transportability of the sediments over the continental domain
            for one or multiple lithologies.
        transportability_mar : float or array-like (m/time)
            The transportability of the sediments over the marine domain for one
            or multiple lithologies.
        wave_base : float (m)
            The wave base, below which weathering decreases exponentially.
        critical_flux : float or array-like (m3/time)
            Critical sediment flux to start displace sediments in the stream
            power law.
        porosity : float or array-like (-)
            The porosity of the sediments at the time of deposition for one or
            multiple lithologies. When computing the active layer, this porosity
            is used unless the field 'sediment__porosity' is being tracked in
            the stratigraphy.
        max_erosion_rate_sed : float (m/time), optional
            The maximum erosion rate of the sediments. If None, all the sediments
            may be eroded in a single time step. The erosion rate defines the
            thickness of the active layer of the sediments if `active_layer_rate_sed`
            is None.
        active_layer_rate_sed : float (m/time), optional
            The rate of formation of the active layer for sediments, which is used
            to determine the composition of the transported sediments. By default,
            it is set by the maximum erosion rate of the sediments.
        bedrock_composition : float or array-like (-)
            The composition of the material is added to the StackedLayers from
            the bedrock.
        max_erosion_rate_br : float (m/time)
            The maximum erosion rate of the bedrock. The erosion rate defines the
            thickness of the active layer of the bedrock if `active_layer_rate_br`
            is None.
        active_layer_rate_br : float (m/time), optional
            The rate of formation of the active layer for the bedrock, which is
            used to determine the composition of the transported sediments. By
            default, it is set by the maximum erosion rate of the bedrock.
        exponent_discharge : float (-)
            The exponent for the water discharge.
        exponent_slope : float (-)
            The exponent for the slope.
        ref_water_flux : float or string (m3/time), optional
            The reference water flux by which the water discharge is normalized.
            If a float, that value is used at each time step; if 'max', the
            maximum value of discharge at each time step is used.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        self._flow_receivers = grid.at_node["flow__receiver_node"][..., np.newaxis]
        n_receivers = self._flow_receivers.shape[1]

        super().__init__(
            grid=grid,
            number_of_neighbors=n_receivers,
            diffusivity_cont=transportability_cont,
            diffusivity_mar=transportability_mar,
            wave_base=wave_base,
            critical_flux=critical_flux,
            porosity=porosity,
            max_erosion_rate_sed=max_erosion_rate_sed,
            active_layer_rate_sed=active_layer_rate_sed,
            bedrock_composition=bedrock_composition,
            max_erosion_rate_br=max_erosion_rate_br,
            active_layer_rate_br=active_layer_rate_br,
            exponent_discharge=exponent_discharge,
            exponent_slope=exponent_slope,
            ref_water_flux=ref_water_flux,
            fields_to_track=fields_to_track,
        )

    def _calculate_sediment_outflux(self, dt):
        """
        Calculates the sediment outflux for multiple lithologies.
        """
        core_nodes = self._grid.core_nodes
        cell_area = self._grid.cell_area_at_node[:, np.newaxis, np.newaxis]

        self._calculate_sediment_diffusivity()
        self._calculate_active_layer_composition(dt)

        self._sediment_outflux[:] = (
            self._K_sed
            * self._active_layer_composition
            * (self._water_flux * self._flow_proportions) ** self._m
            * self._slope**self._n
        )
        self._critical_rate[core_nodes] = self._critical_flux / cell_area[core_nodes]
        self._ratio_critical_outflux[:] = 0.0
        np.divide(
            self._sediment_outflux,
            self._critical_rate,
            out=self._ratio_critical_outflux,
            where=self._critical_rate != 0,
        )
        self._sediment_outflux[:] -= self._critical_rate * (
            1.0 - np.exp(-self._ratio_critical_outflux)
        )
        self._sediment_outflux[:] *= cell_area

    def run_one_step(self, dt, update_compatible=False, update=False):
        """Run the router for one timestep, dt.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        update_compatible : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer at the top of the stack only
            if the new layer is compatible with the existing layer.
        update : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer.
        """
        cell_area = self._grid.cell_area_at_node[:, np.newaxis]

        self._normalize_water_flux()

        # Here we merge fluxes from the sediments and the bedrock together,
        # assuming that weathered bedrock is perfectly equivalent to sediments,
        # including in terms of porosity.
        self._calculate_sediment_outflux(dt)
        if (
            self._max_erosion_rate_sed != self._active_layer_rate_sed
            or self.max_erosion_rate_br != self._active_layer_rate_br
        ):
            self._calculate_active_layer(
                self._max_erosion_rate_sed * dt, self.max_erosion_rate_br * dt
            )
        self._max_sediment_outflux[:] = cell_area * self._active_layer[:, 0] / dt

        self._sediment_influx[:] = self._sediment_input
        calculate_sediment_influx(
            self._node_order,
            self._flow_receivers[..., 0],
            self._sediment_influx,
            self._sediment_outflux,
            self._max_sediment_outflux,
        )

        self._apply_fluxes(dt, update_compatible, update)



"""Utils"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np


################################################################################
# Array manipulation


def convert_to_array(x):
    """
    Converts a scalar or an array-like to a NumPy array.
    """
    return np.array([x]) if np.isscalar(x) else np.asarray(x)


def match_array_lengths(*arrays):
    """
    Duplicates the item in one or several arrays to match the length of the
    longest array.
    """
    lengths_over_one = [len(a) for a in arrays if len(a) > 1]
    if not lengths_over_one:
        return arrays

    _arrays = []
    for i, array in enumerate(arrays):
        if len(array) == 1:
            _arrays.append(np.full(lengths_over_one[0], array[0], dtype=array.dtype))
        else:
            _arrays.append(array)

    return _arrays


def reshape_to_match(x, shape):
    """
    Reshapes an array to match the dimensions in a shape.
    """
    new_shape = np.empty(len(shape), dtype=int)
    for i, size in enumerate(shape):
        try:
            j = x.shape.index(size)
            new_shape[i] = x.shape[j]
        except:
            new_shape[i] = 1

    return x.reshape(new_shape)


################################################################################
# Field manipulation


def format_fields_to_track(fields_to_track):
    """
    Formats some grid fields to track.
    """
    if fields_to_track is None:
        return []
    elif isinstance(fields_to_track, str):
        return [fields_to_track]
    else:
        return list(fields_to_track)



"""Calculate multiple flux divergence on a raster grid"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np

from .cfuncs import _calc_mult_flux_div_at_node


################################################################################
# Flux divergence


def calc_mult_flux_div_at_node(grid, unit_flux, out=None):
    """Calculate the divergences of link-based fluxes at nodes.

    Given a flux per unit width across each face in the grid, calculate the net
    outflux (or influx, if negative) divided by cell area, at each node (zero
    or "out" value for nodes without cells).

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    unit_flux : ndarray
        Flux per unit width along links for multiple classes treated in parallel
        (x number of links, y number of classes).

    Returns
    -------
    ndarray (x number of nodes, y number of classes)
        Flux divergences at nodes.
    """
    if unit_flux.shape[0] != grid.number_of_links:
        raise ValueError("Parameter unit_flux must be num links " "long")
    if out is None:
        out = np.zeros((grid.number_of_nodes, unit_flux.shape[1]))
    elif out.shape[0] != grid.number_of_nodes:
        raise ValueError("output buffer length mismatch with number of nodes")

    out.reshape(*grid.shape, unit_flux.shape[1])[:, (0, -1)] = 0.0
    out.reshape(*grid.shape, unit_flux.shape[1])[(0, -1), :] = 0.0

    _calc_mult_flux_div_at_node(grid.shape, (grid.dx, grid.dy), unit_flux, out)

    return out



"""Sediment compactor"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np
from landlab import Component

from ...utils import convert_to_array, format_fields_to_track


################################################################################
# Component


class SedimentCompactor(Component):
    """Compaction of sediments in a StackedLayers.

    Warning: This component is slow, especially when the number of layers increases.

    References
    ----------
    Granjeon, D. (1996)
        Modélisation stratigraphique déterministe: Conception et applications d'un modèle diffusif 3D multilithologique
        https://theses.hal.science/tel-00648827
    Salles, T., Mallard, C., & Zahirovic, S. (2020)
        goSPL: global scalable paleo landscape evolution
        https://doi.org/10.21105/joss.02804
    """

    _name = "SedimentCompactor"

    _unit_agnostic = True

    _info = {
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "sediment__porosity": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Porosity of the sediments at the surface",
        },
    }

    def __init__(
        self,
        grid,
        initial_porosity=0.5,
        efolding_thickness=1000.0,
        fields_to_track=None,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        initial_porosity : float or array-like (m/time)
            The initial porosity of the sediments for one or multiple lithologies.
        efolding_thickness : float or array-like (m/time)
            The e-folding sediment thickness for one or multiple lithologies.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(grid)

        # Parameters
        self.initial_porosity = convert_to_array(initial_porosity)
        if "sediment__porosity" not in grid.at_node:
            _ = grid.add_field(
                "sediment__porosity",
                np.full(
                    (grid.number_of_nodes, len(self.initial_porosity)),
                    self.initial_porosity,
                ),
            )
        self.efolding_thickness = convert_to_array(efolding_thickness)
        self.fields_to_track = format_fields_to_track(fields_to_track)
        if "sediment__porosity" not in self.fields_to_track:
            self.fields_to_track.append("sediment__porosity")

        # Physical fields
        self._topography = grid.at_node["topographic__elevation"]
        self._stratigraphy = grid.stacked_layers

        # Fields for compaction
        self._initial_thickness = np.zeros(self._stratigraphy.number_of_stacks)

    def run_one_step(self):
        """Run the compactor for one timestep."""
        core_nodes = self._grid.core_nodes

        self._initial_thickness[:] = self._stratigraphy.thickness

        _depth = self._stratigraphy.z[..., np.newaxis]  # This is awfully slow
        _depth[1:] = _depth[1:] - (_depth[1:] - _depth[:-1]) / 2.0
        _depth[0] /= 2.0
        porosity = self.initial_porosity * np.exp(
            -_depth / self.efolding_thickness
        )  # This is awfully slow
        porosity = np.minimum(self._stratigraphy["sediment__porosity"], porosity)

        self._stratigraphy["_dz"][:] *= (
            1.0 - self._stratigraphy["sediment__porosity"]
        ) / (1.0 - porosity)
        self._stratigraphy["sediment__porosity"][:] = porosity
        self._topography[core_nodes] -= (
            self._initial_thickness - self._stratigraphy.thickness
        )



"""Cython functions for stream power"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


cimport cython
from libc.stdint cimport int8_t

# https://cython.readthedocs.io/en/stable/src/userguide/fusedtypes.html
ctypedef fused id_t:
    cython.integral
    long long


################################################################################
# Functions


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void calculate_flux_limiter(
    const id_t [:] node_order,
    const id_t [:, :] neighbors,
    const id_t [:, :] links_to_neighbors,
    const int8_t [:, :] link_dirs_at_node,
    const cython.floating [:, :] sediment_flux_at_links,
    const cython.floating [:, :] max_sediment_flux,
    const cython.floating [:, :] sediment_input,
    cython.floating [:, :] flux_limiter,
) noexcept nogil:
    """Calculates the flux limiter."""
    cdef unsigned int n_nodes = node_order.shape[0]
    cdef unsigned int n_neighbors = neighbors.shape[1]
    cdef unsigned int n_sediments = sediment_input.shape[1]
    cdef unsigned int node, i, j, k
    cdef int neighbor, link
    cdef double total_outflux, max_outflux, outflux

    # Iterate top to bottom through the nodes, update sediment out- and influx.
    # Because calculation of the outflux requires the influx, this operation
    # must be done in an upstream to downstream loop, and cannot be vectorized.
    for i in range(n_nodes - 1, -1, -1):

        # Choose the node id
        node = node_order[i]

        # For each sediment class...
        for k in range(n_sediments):

            # Compute the available sediments, i.e., the maximum ouflux, and the
            # total outflux.
            max_outflux = 0.
            total_outflux = 0.
            for j in range(n_neighbors):
                link = links_to_neighbors[node, j]
                neighbor = neighbors[node, j]
                outflux = sediment_flux_at_links[link, k]*link_dirs_at_node[node, j]
                if outflux <= 0.:
                    total_outflux -= outflux
                else:
                    max_outflux += flux_limiter[neighbor, k]*outflux
            max_outflux += max_sediment_flux[node, k] + sediment_input[node, k]

            # Compute the flux limiter
            if total_outflux > 0. and max_outflux/total_outflux < 1.:
                flux_limiter[node, k] = max_outflux/total_outflux
            else:
                flux_limiter[node, k] = 1.


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void calculate_sediment_influx(
    const id_t [:] node_order,
    const id_t [:, :] flow_receivers,
    cython.floating [:, :] sediment_influx,
    cython.floating [:, :, :] sediment_outflux,
    const cython.floating [:, :] max_sediment_outflux,
) noexcept nogil:
    """Calculates sediment influx."""
    cdef unsigned int n_nodes = node_order.shape[0]
    cdef unsigned int n_receivers = flow_receivers.shape[1]
    cdef unsigned int n_sediments = sediment_outflux.shape[2]
    cdef unsigned int node, i, j, k
    cdef double total_outflux, max_outflux
    cdef double limiter

    # Iterate top to bottom through the nodes, update sediment out- and influx.
    # Because calculation of the outflux requires the influx, this operation
    # must be done in an upstream to downstream loop, and cannot be vectorized.
    for i in range(n_nodes - 1, -1, -1):

        # Choose the node id
        node = node_order[i]

        # For each sediment class...
        for k in range(n_sediments):

            # Compute the available sediments, i.e., the maximum ouflux
            max_outflux = sediment_influx[node, k] + max_sediment_outflux[node, k]
            if max_outflux > 0.:
                # Compute the total outflux
                total_outflux = 0.
                for j in range(n_receivers):
                    total_outflux += sediment_outflux[node, j, k]
                if total_outflux > 0.:
                    # Determine by how much the sediment outflux needs to be decreased
                    limiter = max_outflux/total_outflux
                    if limiter < 1.:
                        # Update the sediment outflux
                        for j in range(n_receivers):
                            sediment_outflux[node, j, k] *= limiter
                    # Add the outflux to the influx of the downstream node(s)
                    for j in range(n_receivers):
                        # TODO: Check this, it's not in the Landlab components, but it seems that it can fail otherwise
                        if flow_receivers[node, j] > -1:
                            sediment_influx[flow_receivers[node, j], k] += sediment_outflux[node, j, k]
            else:
                for j in range(n_receivers):
                    sediment_outflux[node, j, k] = 0.


from libc.math cimport exp
from libc.math cimport log
from libc.math cimport isinf


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void calculate_sediment_fluxes(
    const id_t [:] node_order,
    const cython.floating [:] cell_area,
    const id_t [:, :] flow_receivers,
    const cython.floating [:] water_flux,
    const cython.floating [:, :] flow_proportions,
    cython.floating [:, :] sediment_influx,
    cython.floating [:, :, :] sediment_outflux,
    const cython.floating [:, :] settling_velocity,
    cython.floating [:, :, :] erosion_flux_sed,
    cython.floating [:, :, :] erosion_flux_br,
    cython.floating [:, :] max_sediment_outflux,
    const cython.floating [:] porosity,
) noexcept nogil:
    """Calculates sediment influx and outflux."""
    cdef unsigned int n_nodes = node_order.shape[0]
    cdef unsigned int n_receivers = flow_receivers.shape[1]
    cdef unsigned int n_sediments = sediment_outflux.shape[2]
    cdef unsigned int node, i, j, k
    cdef double total_outflux, max_outflux
    cdef double limiter

    # Iterate top to bottom through the nodes, update sediment out- and influx.
    # Because calculation of the outflux requires the influx, this operation
    # must be done in an upstream to downstream loop, and cannot be vectorized.
    for i in range(n_nodes - 1, -1, -1):

        # Choose the node id
        node = node_order[i]

        # For each sediment class...
        for k in range(n_sediments):

            # Compute the available sediments, i.e., the maximum ouflux
            max_outflux = sediment_influx[node, k] + max_sediment_outflux[node, k]
            for j in range(n_receivers):
                max_outflux += erosion_flux_br[node, j, k]
            if max_outflux > 0.:
                # Compute the sediment outflux
                for j in range(n_receivers):
                    if flow_proportions[node, j]*water_flux[node] > 0.:
                        # This applies the formula for sediment outlfux from
                        # Shobe et al. (2017) to each receiver independantly by
                        # assuming that the sediment influx is divided between
                        # each receiver in the same proportions than the water
                        # flux, which might not be the correct approach
                        sediment_outflux[node, j, k] = (
                            flow_proportions[node, j]*sediment_influx[node, k]
                            + erosion_flux_sed[node, j, k]
                            + erosion_flux_br[node, j, k]*(1. - porosity[k])
                        )
                        sediment_outflux[node, j, k] /= 1. + settling_velocity[node, k]*cell_area[node]/(flow_proportions[node, j]*water_flux[node])
                # Compute the total outflux
                total_outflux = 0.
                for j in range(n_receivers):
                    total_outflux += sediment_outflux[node, j, k]
                if total_outflux > 0.:
                    # Determine by how much the sediment outflux needs to be decreased
                    limiter = max_outflux/total_outflux
                    if limiter < 1.:
                        # Update the sediment outflux
                        for j in range(n_receivers):
                            sediment_outflux[node, j, k] *= limiter

                    # Add the outflux to the influx of the downstream node(s)
                    for j in range(n_receivers):
                        # TODO: Check this, it's not in the Landlab components, but it seems that it can fail otherwise
                        if flow_receivers[node, j] > -1:
                            sediment_influx[flow_receivers[node, j], k] += sediment_outflux[node, j, k]



"""Gravity-driven diffuser"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np

from ...utils import convert_to_array
from .._base import _BaseDiffuser
from ..stream_power.cfuncs import calculate_flux_limiter


################################################################################
# Component


class GravityDrivenDiffuser(_BaseDiffuser):
    """Gravity-driven diffusion of a Landlab field in continental and marine domains
    based on a finite-volume scheme. The diffusion can be linear (the default) or
    non-linear (when the critical-anlge parameters are defined) following Roering
    et al. (1999).

    This component is not a re-implementation of Gervais (2004), but it is
    heavily based on it. Instead of a implicit or semi-implicit approach
    starting from updating the topography, it uses a fully explicit approach
    that ends with updating the topography.

    References
    ----------
    Gervais, V. (2004)
        Étude et Simulation d'un Modèle Stratigraphique Multi-Lithologique sous Contrainte de Taux d'Érosion Maximal
        https://theses.hal.science/tel-01445562/
    Roering, J. J., Kirchner, J. W., & Dietrich, W. E. (1999)
        Evidence for nonlinear, diffusive sediment transport on hillslopes and implications for landscape morphology
        https://doi.org/10.1029/1998WR900090
    """

    _name = "GravityDrivenDiffuser"

    def __init__(
        self,
        grid,
        diffusivity_cont=0.01,
        diffusivity_mar=0.001,
        critical_angle_cont=None,
        critical_angle_mar=None,
        wave_base=20.0,
        porosity=0.0,
        max_erosion_rate_sed=0.01,
        active_layer_rate_sed=None,
        bedrock_composition=1.0,
        max_erosion_rate_br=0.01,
        active_layer_rate_br=None,
        exponent_slope=1.0,
        fields_to_track=None,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        diffusivity_cont : float or array-like (m2/time)
            The diffusivity of the sediments over the continental domain for one
            or multiple lithologies.
        diffusivity_mar : float or array-like (m2/time)
            The diffusivity of the sediments over the marine domain for one or
            multiple lithologies.
        critical_angle_cont : float or array-like (degree), optional
            The critical angle in the continental domain for one or multiple
            lithologies above which sediments move downslope by mass wasting.
        critical_angle_mar : float or array-like (degree), optional
            The critical angle in the continental domain for one or multiple
             ithologies above which sediments move downslope by mass wasting.
        wave_base : float (m)
            The wave base, below which weathering decreases exponentially.
        porosity : float or array-like (-)
            The porosity of the sediments at the time of deposition for one or
            multiple lithologies. When computing the active layer, this porosity
            is used unless the field 'sediment__porosity' is being tracked in
            the stratigraphy.
        max_erosion_rate_sed : float (m/time), optional
            The maximum erosion rate of the sediments. If None, all the sediments
            may be eroded in a single time step. The erosion rate defines the
            thickness of the active layer of the sediments if `active_layer_rate`
            is None.
        active_layer_rate_sed : float (m/time), optional
            The rate of formation of the active layer for sediments, which is used
            to determine the composition of the transported sediments. By default,
            it is set by the maximum erosion rate of the sediments.
        bedrock_composition : float or array-like (-)
            The composition of the material is added to the StackedLayers from
            the bedrock.
        max_erosion_rate_br : float (m/time)
            The maximum erosion rate of the bedrock. The erosion rate defines the
            thickness of the active layer of the bedrock if `active_layer_rate`
            is None.
        active_layer_rate_br : float (m/time), optional
            The rate of formation of the active layer for the bedrock, which is
            used to determine the composition of the transported sediments. By
            default, it is set by the maximum erosion rate of the bedrock.
        exponent_slope : float (-)
            The exponent for the slope.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(
            grid,
            diffusivity_cont,
            diffusivity_mar,
            wave_base,
            porosity,
            max_erosion_rate_sed,
            active_layer_rate_sed,
            bedrock_composition,
            max_erosion_rate_br,
            active_layer_rate_br,
            exponent_slope,
            fields_to_track,
        )

        # Parameters
        if critical_angle_cont is None:
            self._critical_slope_cont = None
        else:
            self._critical_slope_cont = convert_to_array(critical_angle_cont)
            if np.all(self._critical_slope_cont == 0.0):
                self._critical_slope_cont = None
            else:
                self._critical_slope_cont = np.tan(
                    np.deg2rad(self._critical_slope_cont)
                )
        if critical_angle_mar is None:
            self._critical_slope_mar = None
        else:
            self._critical_slope_mar = convert_to_array(critical_angle_mar)
            if np.all(self._critical_slope_mar == 0.0):
                self._critical_slope_mar = None
            else:
                self._critical_slope_mar = np.tan(np.deg2rad(self._critical_slope_mar))

        # Physical fields
        n_links = grid.number_of_links
        self._bathymetry_at_links = np.zeros(n_links)

        # Fields for sediment fluxes
        n_nodes = grid.number_of_nodes
        n_sediments = self._stratigraphy.number_of_classes
        self._node_order = np.zeros(n_nodes, dtype=int)
        self._K_sed_at_links = np.zeros((n_links, n_sediments))
        self._slope_at_links = np.zeros((n_links, 1))
        if (
            self._critical_slope_cont is not None
            and self._critical_slope_mar is not None
        ):
            self._thres_slope_at_links = np.zeros((n_links, n_sediments))
            self._critical_slope_at_links = np.zeros((n_links, n_sediments))
        self._active_layer_composition_at_links = np.zeros((n_links, n_sediments))
        self._sediment_flux_at_links = np.zeros((n_links, n_sediments))
        self._sediment_flux = np.zeros((n_links, n_sediments))
        self._sediment_rate = np.zeros((n_nodes, n_sediments))
        self._max_sediment_flux = np.zeros((n_nodes, n_sediments))
        self._sediment_input = np.zeros((n_nodes, n_sediments))
        self._flux_limiter = np.zeros((n_nodes, n_sediments))
        self._flux_limiter_at_links = np.zeros((n_links, n_sediments))

    def _adjust_sediment_diffusivity(self):
        """
        Ajusts the sediment diffusivity based on the slope over the continental
        and marine domains.
        """
        self._grid.map_mean_of_link_nodes_to_link(
            self._bathymetry[:, 0], out=self._bathymetry_at_links
        )

        self._critical_slope_at_links[self._bathymetry_at_links == 0.0] = (
            self._critical_slope_cont
        )
        self._critical_slope_at_links[self._bathymetry_at_links > 0.0] = (
            self._critical_slope_mar
        )

        self._thres_slope_at_links[:] = np.where(
            np.abs(self._slope_at_links) >= self._critical_slope_at_links,
            self._critical_slope_at_links - 1e-12,
            self._slope_at_links,
        )

        self._K_sed_at_links /= (
            1.0 - (self._thres_slope_at_links / self._critical_slope_at_links) ** 2
        )

    def _calculate_sediment_flux(self, dt):
        """
        Calculates the sediment fluxes for multiple lithologies.
        """
        self._slope_at_links[self._grid.active_links, 0] = self._grid.calc_grad_at_link(
            self._topography
        )[self._grid.active_links]
        self._calculate_active_layer_composition(dt)
        self._calculate_sediment_diffusivity()

        self._grid.map_mean_of_link_nodes_to_link(
            self._K_sed[:, 0], out=self._K_sed_at_links
        )
        self._grid.map_value_at_max_node_to_link(
            self._topography[:, np.newaxis],
            self._active_layer_composition[:, 0],
            out=self._active_layer_composition_at_links,
        )

        if (
            self._critical_slope_cont is not None
            and self._critical_slope_mar is not None
        ):
            self._adjust_sediment_diffusivity()

        self._sediment_flux_at_links[:] = (
            -self._K_sed_at_links
            * np.sign(self._slope_at_links)
            * self._active_layer_composition_at_links
            * np.abs(self._slope_at_links) ** self._n
        )

    def _threshold_sediment_flux(self, dt):
        """
        Thresholds the sediment fluxes to not exceed a maximum erosion rate.
        """
        cell_area = self._grid.cell_area_at_node[:, np.newaxis]

        # Here we merge fluxes from the sediments and the bedrock together,
        # assuming that weathered bedrock is perfectly equivalent to sediments,
        # including in terms of porosity.
        if (
            self._max_erosion_rate_sed != self._active_layer_rate_sed
            or self.max_erosion_rate_br != self._active_layer_rate_br
        ):
            self._calculate_active_layer(
                self._max_erosion_rate_sed * dt, self.max_erosion_rate_br * dt
            )
        self._max_sediment_flux[:] = cell_area * self._active_layer[:, 0] / dt

        self._node_order[:] = np.argsort(self._topography)
        self._sediment_flux[self._grid.link_at_face] = (
            self._sediment_flux_at_links[self._grid.link_at_face]
            * self._grid.length_of_face[:, np.newaxis]
        )
        calculate_flux_limiter(
            self._node_order,
            self._grid.active_adjacent_nodes_at_node,
            self._grid.links_at_node,
            self._grid.link_dirs_at_node,
            self._sediment_flux,
            self._max_sediment_flux,
            self._sediment_input,
            self._flux_limiter,
        )
        self._grid.map_value_at_max_node_to_link(
            self._topography[:, np.newaxis],
            self._flux_limiter,
            out=self._flux_limiter_at_links,
        )

        self._sediment_flux_at_links *= self._flux_limiter_at_links

    def run_one_step(self, dt, update_compatible=False, update=False):
        """Run the diffuser for one timestep, dt.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        update_compatible : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer at the top of the stack only
            if the new layer is compatible with the existing layer.
        update : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer.
        """
        core_nodes = self._grid.core_nodes

        self._calculate_sediment_flux(dt)
        self._threshold_sediment_flux(dt)
        self._grid.calc_mult_flux_div_at_node(
            self._sediment_flux_at_links, out=self._sediment_rate
        )

        self._sediment_thickness[core_nodes] = (
            -self._sediment_rate[core_nodes] * dt / (1.0 - self._porosity)
        )

        self._update_physical_fields(dt, update_compatible, update)



"""Grids with stacked layers"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import importlib
import os
from landlab.core.utils import get_functions_from_module, add_functions_to_class
from landlab import RasterModelGrid as _RasterModelGrid
from landlab import VoronoiDelaunayGrid as _VoronoiDelaunayGrid
from landlab import FramedVoronoiGrid as _FramedVoronoiGrid
from landlab import HexModelGrid as _HexModelGrid

from ..layers import StackedLayersMixIn
from ..plot.layers import RasterModelGridLayerPlotterMixIn


################################################################################
# Adding member function (from Landlab)


def add_module_functions_to_class(cls, module, pattern=None, exclude=None):
    """Add functions from a module to a class as methods.

    Parameters
    ----------
    cls : class
        A class.
    module : module
        An instance of a module.
    pattern : str, optional
        Only get functions whose name match a regular expression.
    exclude : str, optional
        Only get functions whose name exclude the regular expression.

    *Note* if both pattern and exclude are provided both conditions must be met.
    """
    (module, _) = os.path.splitext(os.path.basename(module))

    mod = importlib.import_module("." + module, package="stratigrapy.grid")

    funcs = get_functions_from_module(mod, pattern=pattern, exclude=exclude)
    add_functions_to_class(cls, funcs)


################################################################################
# Raster


class RasterModelGrid(
    StackedLayersMixIn, RasterModelGridLayerPlotterMixIn, _RasterModelGrid
):
    """A 2D uniform rectilinear grid."""

    def __init__(
        self,
        shape,
        xy_spacing=1.0,
        xy_of_lower_left=(0.0, 0.0),
        xy_of_reference=(0.0, 0.0),
        xy_axis_name=("x", "y"),
        xy_axis_units="-",
        bc=None,
        number_of_classes=1,
        initial_allocation=1,
        new_allocation=1,
        number_of_layers_to_fuse=1,
        number_of_top_layers=1,
        fuse_continuously=True,
        remove_empty_layers=False,
    ):
        """Create a 2D grid with equal spacing.

        Optionally takes numbers of rows and columns and cell size as
        inputs. If this are given, calls initialize() to set up the grid.
        At the moment, num_rows and num_cols MUST be specified. Both must be
        >=3 to allow correct automated setup of boundary conditions.

        Parameters
        ----------
        shape : tuple of int
            Shape of the grid in nodes as (nrows, ncols).
        xy_spacing : tuple or float, optional
            dx and dy spacing. Either provided as a float or a
            (dx, dy) tuple.
        xy_of_lower_left: tuple, optional
            (x, y) coordinates of the lower left corner.
        xy_of_reference : tuple, optional
            Coordinate value in projected space of the reference point,
            `xy_of_lower_left`. Default is (0., 0.)
        xy_axis_name: tuple of str
            Name to use for each axis.
        xy_axis_units: tuple of str, or str
            Units for coordinates of each axis.
        bc : dict, optional
            Edge boundary conditions.
        number_of_classes : int, optional
            Number of classes of material (e.g., different lithologies or grain
            sizes) making the layers of the StackedLayers.
        initial_allocation : int or array-like, optional
            Number of layers initially pre-allocated to speed up adding new layers.
            If an int, the layers are pre-allocated at the end; it an array-like,
            the first element corresponds to layers pre-allocated at the start,
            the second to layers pre-allocated at the end.
        new_allocation : int or array-like, optional
            Number of layers pre-allocated when adding new layers to speed up the
            process. If an array-like, different number of layers are pre-allocated
            at the start and at the end of the stacks.
        number_of_layers_to_fuse : int, optional
            Number of layers to merge together when calling `fuse`.
        number_of_top_layers : int, optional
            Number of layers at the top of the stack to ignore when calling `fuse`.
        fuse_continuously : bool, optional
            If True, fuse the layers as they are added; otherwise, wait until the
            number of unfused layers reaches `number_of_layers_to_fuse` (excluding
            the top layers) to fuse them.
        remove_empty_layers : bool, optional
            If True, remove the layers without any deposits at the top of the stack
            when adding a layer; otherwise, keep all the layers.

        Returns
        -------
        RasterModelGrid
            A newly-created grid.
        """
        _RasterModelGrid.__init__(
            self,
            shape,
            xy_spacing,
            xy_of_lower_left,
            xy_of_reference,
            xy_axis_name,
            xy_axis_units,
            bc,
        )
        StackedLayersMixIn.__init__(
            self,
            number_of_classes,
            initial_allocation,
            new_allocation,
            number_of_layers_to_fuse,
            number_of_top_layers,
            fuse_continuously,
            remove_empty_layers,
        )


add_module_functions_to_class(RasterModelGrid, "raster_divergence.py", pattern="calc_*")


################################################################################
# Voronoi


class VoronoiDelaunayGrid(StackedLayersMixIn, _VoronoiDelaunayGrid):
    """This inherited class implements an unstructured grid in which cells are
    Voronoi polygons and nodes are connected by a Delaunay triangulation. Uses
    scipy.spatial module to build the triangulation.

    Create an unstructured grid from points whose coordinates are given
    by the arrays *x*, *y*.
    """

    def __init__(
        self,
        x=None,
        y=None,
        reorient_links=True,
        xy_of_reference=(0.0, 0.0),
        xy_axis_name=("x", "y"),
        xy_axis_units="-",
        number_of_classes=1,
        initial_allocation=1,
        new_allocation=1,
        number_of_layers_to_fuse=1,
        number_of_top_layers=1,
        fuse_continuously=True,
        remove_empty_layers=False,
    ):
        """Create a Voronoi Delaunay grid from a set of points.

        Create an unstructured grid from points whose coordinates are given
        by the arrays *x*, *y*.

        Parameters
        ----------
        x : array_like
            x-coordinate of points
        y : array_like
            y-coordinate of points
        reorient_links (optional) : bool
            whether to point all links to the upper-right quadrant
        xy_of_reference : tuple, optional
            Coordinate value in projected space of (0., 0.)
            Default is (0., 0.)
        number_of_classes : int, optional
            Number of classes of material (e.g., different lithologies or grain
            sizes) making the layers of the StackedLayers.
        initial_allocation : int or array-like, optional
            Number of layers initially pre-allocated to speed up adding new layers.
            If an int, the layers are pre-allocated at the end; it an array-like,
            the first element corresponds to layers pre-allocated at the start,
            the second to layers pre-allocated at the end.
        new_allocation : int or array-like, optional
            Number of layers pre-allocated when adding new layers to speed up the
            process. If an array-like, different number of layers are pre-allocated
            at the start and at the end of the stacks.
        number_of_layers_to_fuse : int, optional
            Number of layers to merge together when calling `fuse`.
        number_of_top_layers : int, optional
            Number of layers at the top of the stack to ignore when calling `fuse`.
        fuse_continuously : bool, optional
            If True, fuse the layers as they are added; otherwise, wait until the
            number of unfused layers reaches `number_of_layers_to_fuse` (excluding
            the top layers) to fuse them.
        remove_empty_layers : bool, optional
            If True, remove the layers without any deposits at the top of the stack
            when adding a layer; otherwise, keep all the layers.

        Returns
        -------
        VoronoiDelaunayGrid
            A newly-created grid.
        """
        _VoronoiDelaunayGrid.__init__(
            self, x, y, reorient_links, xy_of_reference, xy_axis_name, xy_axis_units
        )
        StackedLayersMixIn.__init__(
            self,
            number_of_classes,
            initial_allocation,
            new_allocation,
            number_of_layers_to_fuse,
            number_of_top_layers,
            fuse_continuously,
            remove_empty_layers,
        )


add_module_functions_to_class(VoronoiDelaunayGrid, "divergence.py", pattern="calc_*")


class FramedVoronoiGrid(StackedLayersMixIn, _FramedVoronoiGrid):
    """A grid of Voronoi Delaunay cells with a structured perimeter layout.

    This inherited class implements a irregular 2D grid with Voronoi Delaunay cells and
    irregular patches. It is a special type of :class:`~.VoronoiDelaunayGrid` grid in which
    the initial set of points is arranged in a fixed lattice (e.g. like a
    :class:`~.RasterModelGrid`), named here "layout", and the core points are
    then moved a random distance from their initial positions, bounded by a user-supplied
    threshold.
    """

    def __init__(
        self,
        shape,
        xy_spacing=(1.0, 1.0),
        xy_of_lower_left=(0.0, 0.0),
        xy_min_spacing=(0.5, 0.5),
        seed=200,
        xy_of_reference=(0.0, 0.0),
        xy_axis_name=("x", "y"),
        xy_axis_units="-",
        number_of_classes=1,
        initial_allocation=1,
        new_allocation=1,
        number_of_layers_to_fuse=1,
        number_of_top_layers=1,
        fuse_continuously=True,
        remove_empty_layers=False,
    ):
        """Create a grid of voronoi cells with a structured perimeter.

        Create an irregular 2D grid with voronoi cells and triangular patches.
        It is a special type of :class:`~.VoronoiDelaunayGrid` in which the initial set
        of points is arranged in a regular lattice determined by the parameters
        *shape*, and *xy_spacing*. The coordinates of
        the core points are then randomly moved while the perimeter points
        remaining fixed, in a way determined by the parameters *xy_min_spacing*, and
        *seed*.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of nodes.
        xy_spacing : float or tuple of float, optional
            Node spacing along x and y coordinates. If ``float``, same spacing at *x* and *y*.
        xy_of_lower_left : tuple, optional
            Minimum *x*-of-node and *y*-of-node values. Depending on the grid,
            there may not be a node at this coordinate.
        xy_min_spacing: float or tuple of float, optional
            Final minimal spacing between nodes. Random moves of the core nodes
            from their initial positions cannot be above this threshold:
            ``(xy_spacing - xy_min_spacing) / 2``
            If ``float``, same minimal spacing for *x* and *y*.
        seed: int, optional
            Seed used to generate the random *x* and *y* moves.
            When set, controls the pseudo-randomness of moves to ensure
            reproducibility.
            When ``None``, the seed is random and the moves of coordinates are
            completely random.
        xy_of_reference : tuple, optional
            Coordinate value in projected space of the reference point,
            *xy_of_lower_left*.
        xy_axis_name: tuple of str, optional
            *x* and *y* axis names.
        xy_axis_units: str, optional
            *x* and *y* axis units.
        number_of_classes : int, optional
            Number of classes of material (e.g., different lithologies or grain
            sizes) making the layers of the StackedLayers.
        initial_allocation : int or array-like, optional
            Number of layers initially pre-allocated to speed up adding new layers.
            If an int, the layers are pre-allocated at the end; it an array-like,
            the first element corresponds to layers pre-allocated at the start,
            the second to layers pre-allocated at the end.
        new_allocation : int or array-like, optional
            Number of layers pre-allocated when adding new layers to speed up the
            process. If an array-like, different number of layers are pre-allocated
            at the start and at the end of the stacks.
        number_of_layers_to_fuse : int, optional
            Number of layers to merge together when calling `fuse`.
        number_of_top_layers : int, optional
            Number of layers at the top of the stack to ignore when calling `fuse`.
        fuse_continuously : bool, optional
            If True, fuse the layers as they are added; otherwise, wait until the
            number of unfused layers reaches `number_of_layers_to_fuse` (excluding
            the top layers) to fuse them.
        remove_empty_layers : bool, optional
            If True, remove the layers without any deposits at the top of the stack
            when adding a layer; otherwise, keep all the layers.

        Returns
        -------
        FramedVoronoiGrid
            A newly-created grid.
        """
        _FramedVoronoiGrid.__init__(
            self,
            shape,
            xy_spacing,
            xy_of_lower_left,
            xy_min_spacing,
            seed,
            xy_of_reference,
            xy_axis_name,
            xy_axis_units,
        )
        StackedLayersMixIn.__init__(
            self,
            number_of_classes,
            initial_allocation,
            new_allocation,
            number_of_layers_to_fuse,
            number_of_top_layers,
            fuse_continuously,
            remove_empty_layers,
        )


add_module_functions_to_class(FramedVoronoiGrid, "divergence.py", pattern="calc_*")


################################################################################
# Hex


class HexModelGrid(StackedLayersMixIn, _HexModelGrid):
    """A grid of hexagonal cells.

    This inherited class implements a regular 2D grid with hexagonal cells and
    triangular patches. It is a special type of VoronoiDelaunay grid in which
    the initial set of points is arranged in a triangular/hexagonal lattice.
    """

    def __init__(
        self,
        shape,
        spacing=1.0,
        xy_of_lower_left=(0.0, 0.0),
        orientation="horizontal",
        node_layout="hex",
        reorient_links=True,
        xy_of_reference=(0.0, 0.0),
        xy_axis_name=("x", "y"),
        xy_axis_units="-",
        number_of_classes=1,
        initial_allocation=1,
        new_allocation=1,
        number_of_layers_to_fuse=1,
        number_of_top_layers=1,
        fuse_continuously=True,
        remove_empty_layers=False,
    ):
        """Create a grid of hexagonal cells.

        Create a regular 2D grid with hexagonal cells and triangular patches.
        It is a special type of :class:`~.VoronoiModelGrid` in which the initial set
        of points is arranged in a triangular/hexagonal lattice.

        Parameters
        ----------
        shape : tuple of int
            Number of rows and columns of nodes.
        spacing : float, optional
            Node spacing.
        xy_of_lower_left : tuple of float, optional
            Minimum x-of-node and y-of-node values. Depending on the grid
            no node may be present at this coordinate. Default is (0., 0.).
        xy_of_reference : tuple of float, optional
            Coordinate value in projected space of the reference point,
            `xy_of_lower_left`. Default is (0., 0.)
        orientation : str, optional
            One of the 3 cardinal directions in the grid, either 'horizontal'
            (default) or 'vertical'
        node_layout : {"hex", "rect"}
            The grid layout of nodes.
        reorient_links : bool, optional
            Whether or not to re-orient all links to point between -45 deg
            and +135 deg clockwise from "north" (i.e., along y axis). default
            is True.
        number_of_classes : int, optional
            Number of classes of material (e.g., different lithologies or grain
            sizes) making the layers of the StackedLayers.
        initial_allocation : int or array-like, optional
            Number of layers initially pre-allocated to speed up adding new layers.
            If an int, the layers are pre-allocated at the end; it an array-like,
            the first element corresponds to layers pre-allocated at the start,
            the second to layers pre-allocated at the end.
        new_allocation : int or array-like, optional
            Number of layers pre-allocated when adding new layers to speed up the
            process. If an array-like, different number of layers are pre-allocated
            at the start and at the end of the stacks.
        number_of_layers_to_fuse : int, optional
            Number of layers to merge together when calling `fuse`.
        number_of_top_layers : int, optional
            Number of layers at the top of the stack to ignore when calling `fuse`.
        fuse_continuously : bool, optional
            If True, fuse the layers as they are added; otherwise, wait until the
            number of unfused layers reaches `number_of_layers_to_fuse` (excluding
            the top layers) to fuse them.
        remove_empty_layers : bool, optional
            If True, remove the layers without any deposits at the top of the stack
            when adding a layer; otherwise, keep all the layers.

        Returns
        -------
        HexModelGrid
            A newly-created grid.
        """
        _HexModelGrid.__init__(
            self,
            shape,
            spacing,
            xy_of_lower_left,
            orientation,
            node_layout,
            reorient_links,
            xy_of_reference,
            xy_axis_name,
            xy_axis_units,
        )
        StackedLayersMixIn.__init__(
            self,
            number_of_classes,
            initial_allocation,
            new_allocation,
            number_of_layers_to_fuse,
            number_of_top_layers,
            fuse_continuously,
            remove_empty_layers,
        )


add_module_functions_to_class(HexModelGrid, "divergence.py", pattern="calc_*")



"""Cython functions for landsliding"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


cimport cython
from libc.math cimport atan, M_PI

# https://cython.readthedocs.io/en/stable/src/userguide/fusedtypes.html
ctypedef fused id_t:
    cython.integral
    long long


################################################################################
# Functions


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void calculate_sediment_influx(
    const id_t [:] node_order,
    const id_t [:] flow_receivers,
    const cython.floating [:] link_lengths,
    const cython.floating [:] cell_area,
    const cython.floating [:] slope,
    const cython.floating [:, :] repose_angle,
    cython.floating [:, :] sediment_influx,
    cython.floating [:, :] sediment_outflux,
    const double dt,
) noexcept nogil:
    """Calculates sediment influx."""
    cdef unsigned int n_nodes = node_order.shape[0]
    cdef unsigned int n_sediments = sediment_outflux.shape[1]
    cdef unsigned int node, i, k
    cdef double total_sediment_influx, repose_slope, new_slope

    # Iterate top to bottom through the nodes, update sediment out- and influx.
    # Because calculation of the outflux requires the influx, this operation
    # must be done in an upstream to downstream loop, and cannot be vectorized.
    for i in range(n_nodes - 1, -1, -1):

        # Choose the node id
        node = node_order[i]

        # Here we assume that the influx gets transported (if needed) before the
        # outflux, i.e., we use the current slope instead of calculating the
        # new slope after the ouflux is gone. This might be conceptually less
        # correct, but it's numerically slightly easier and in the grand scheme
        # of things I don't think it matters much
        # Compute the slope of repose of the influx
        repose_slope = 0.
        total_sediment_influx = 0.
        for k in range(n_sediments):
            total_sediment_influx += sediment_influx[node, k]
            repose_slope += repose_angle[node, k]*sediment_influx[node, k]
        if total_sediment_influx > 0.:
            repose_slope /= n_sediments*total_sediment_influx
            repose_slope = atan(repose_slope*M_PI/180.)
            # Compute the new depositional slope
            new_slope = slope[node]
            if cell_area[node] > 0. and link_lengths[node] > 0.:
                new_slope += total_sediment_influx*dt/cell_area[node]/link_lengths[node]
            # If the new slope is larger than the slope of repose...
            if new_slope > repose_slope:
                # Add the excess sediments to the outflux
                for k in range(n_sediments):
                    sediment_outflux[node, k] += min(sediment_influx[node, k],
                                                     (sediment_influx[node, k]/total_sediment_influx)*(new_slope - repose_slope)*link_lengths[node]*cell_area[node]/dt)

        # TODO: Check this, it's not in the Landlab components, but it seems that it can fail otherwise
        if flow_receivers[node] > -1:
            # For each sediment class...
            for k in range(n_sediments):
                # Add the outflux to the influx of the downstream node(s)
                sediment_influx[flow_receivers[node], k] += sediment_outflux[node, k]



"""StratigraPy is a Python package for stratigraphic modeling based on Landlab"""

from stratigrapy.grid import RasterModelGrid
from stratigrapy.grid import VoronoiDelaunayGrid
from stratigrapy.grid import FramedVoronoiGrid
from stratigrapy.grid import HexModelGrid

__all__ = [
    "HexModelGrid",
    "RasterModelGrid",
    "FramedVoronoiGrid",
    "VoronoiDelaunayGrid",
]



"""Stacked layers"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import os
import numpy as np
from landlab.layers.eventlayers import (
    _reduce_matrix,
    _valid_keywords_or_raise,
    _allocate_layers_for,
)

from .cfuncs import (
    deposit_or_erode,
    get_surface_index,
    get_superficial_layer,
    get_active_layer,
)


################################################################################
# Base functions


def _broadcast(x, shape):
    """
    Broadcasts a float or an array into an array of shape `shape`.
    """
    try:
        x = x.reshape(shape)
    except (AttributeError, ValueError):
        x = np.broadcast_to(x, shape)
    finally:
        x = np.asarray(x, dtype=float)

    return x


def _deposit_or_erode(layers, bottom_index, top_index, dz):
    """Update the array that contains layers with deposition or erosion."""
    layers = layers.reshape((layers.shape[0], layers.shape[1], -1))
    dz = _broadcast(dz, layers.shape[1:3])

    deposit_or_erode(layers, bottom_index, top_index, dz)


def _get_surface_index(layers, bottom_index, top_index, surface_index):
    """Get index within each stack of the layer at the topographic surface."""
    layers = layers.reshape((layers.shape[0], layers.shape[1], -1))

    get_surface_index(layers, bottom_index, top_index, surface_index)


def _get_superficial_layer(layers, bottom_index, top_index, dz, superficial_layer):
    """Get the thicknesses of all classes in a superficial layer defined by its
    thickness from the surface `dz`.
    """
    layers = layers.reshape((layers.shape[0], layers.shape[1], -1))
    dz = _broadcast(dz, layers.shape[1:2])

    get_superficial_layer(layers, bottom_index, top_index, dz, superficial_layer)


def _get_active_layer(layers, porosity, bottom_index, top_index, dz, active_layer):
    """Get the thicknesses of all classes in an active layer defined by its
    thickness from the surface `dz`.
    """
    layers = layers.reshape((layers.shape[0], layers.shape[1], -1))
    dz = _broadcast(dz, layers.shape[1:2])

    if porosity is not None and porosity.shape == layers.shape:
        get_active_layer(layers, porosity, bottom_index, top_index, dz, active_layer)
    else:
        get_superficial_layer(layers, bottom_index, top_index, dz, active_layer)
        if porosity is not None:
            active_layer *= 1.0 - porosity


class _BlockSlice:
    """Slices that divide a matrix into equally sized blocks."""

    def __init__(self, *args):
        """_BlockSlice([start], stop, [step])"""
        if len(args) > 3:
            raise TypeError(
                f"_BlockSlice expected at most 3 arguments, got {len(args)}"
            )

        self._args = tuple(args)

        self._start = 0
        self._stop = None
        self._step = None

        if len(args) == 1:
            self._stop = args[0]
        elif len(args) == 2:
            self._start, self._stop = args
        elif len(args) == 3:
            self._start, self._stop, self._step = args

        if self._stop is not None and self._stop < self._start:
            raise ValueError(
                "stop ({}) must be greater than start ({})".format(
                    self._stop, self._start
                )
            )

    def __repr__(self):
        return "_BlockSlice({})".format(", ".join([repr(arg) for arg in self._args]))

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def step(self):
        return self._step

    def indices(self, n_rows):
        """Row indices to blocks within a matrix.

        Parameters
        ----------
        n_rows : int
            The number of rows in the matrix.

        Returns
        -------
        (start, stop, step)
            Tuple of (int* that gives the row of the first block, row of the
            last block, and the number of rows in each block.
        """
        start, stop, step = self.start, self.stop, self.step
        if stop is None:
            stop = n_rows

        start, stop, _ = slice(start, stop).indices(n_rows)

        if step is None:
            step = stop - start

        if step != 0 and (stop - start) % step != 0:
            stop = (stop - start) // step * step + start

        return start, stop, step


def resize_array(array, left_new_cap, right_new_cap):
    """Increase the size of an array, leaving room to grow.

    Parameters
    ----------
    array : ndarray
        The array to resize.
    left_new_cap : int
        New capacity at the start of the zero-th dimension of the resized array.
    right_new_cap : int
        New capacity at the end of the zero-th dimension of the resized array.

    Returns
    -------
    ndarray
        Copy of the input array resized.
    """
    left_new_cap = int(left_new_cap)
    right_new_cap = int(right_new_cap)
    if left_new_cap == 0 and right_new_cap == 0:
        return array

    new_allocated = left_new_cap + array.shape[0] + right_new_cap
    larger_array = np.zeros((new_allocated,) + array.shape[1:], dtype=array.dtype)
    _right_new_cap = -right_new_cap if right_new_cap > 0 else None
    larger_array[left_new_cap:_right_new_cap] = array

    return larger_array


################################################################################
# StackedLayersMixIn


class StackedLayersMixIn:
    """MixIn that adds a StackedLayers attribute to a ModelGrid."""

    def __init__(
        self,
        number_of_classes,
        initial_allocation,
        new_allocation,
        number_of_layers_to_fuse,
        number_of_top_layers,
        fuse_continuously,
        remove_empty_layers,
    ):
        """Instantiates the member variables of the StackedLayersMixIn.

        Parameters
        ----------
        number_of_classes : int
            Number of classes of material (e.g., different lithologies or grain
            sizes) making the layers of the StackedLayers.
        initial_allocation : int or array-like, optional
            Number of layers initially pre-allocated to speed up adding new layers.
            If an int, the layers are pre-allocated at the end; it an array-like,
            the first element corresponds to layers pre-allocated at the start, the
            second to layers pre-allocated at the end.
        new_allocation : int or array-like, optional
            Number of layers pre-allocated when adding new layers to speed up the
            process. If an int, the layers are pre-allocated at the end; it an
            array-like, the first element corresponds to layers pre-allocated at the
            start, the second to layers pre-allocated at the end.
        number_of_layers_to_fuse : int, optional
            Number of layers to merge together when calling `fuse`.
        number_of_top_layers : int, optional
            Number of layers at the top of the stack to ignore when calling `fuse`.
        fuse_continuously : bool, optional
            If True, fuse the layers as they are added; otherwise, wait until the
            number of unfused layers reaches `number_of_layers_to_fuse` (excluding
            the top layers) to fuse them.
        remove_empty_layers : bool, optional
            If True, remove the layers without any deposits at the top of the stack
            when adding a layer; otherwise, keep all the layers.
        """
        self.number_of_classes = number_of_classes
        self.initial_allocation = initial_allocation
        self.new_allocation = new_allocation
        self.number_of_layers_to_fuse = number_of_layers_to_fuse
        self.number_of_top_layers = number_of_top_layers
        self.fuse_continuously = fuse_continuously
        self.remove_empty_layers = remove_empty_layers

    @property
    def stacked_layers(self):
        """StackedLayers for each cell."""
        try:
            self._stacked_layers
        except AttributeError:
            self._stacked_layers = StackedLayers(
                self.number_of_cells,
                self.number_of_classes,
                self.initial_allocation,
                self.new_allocation,
                self.number_of_layers_to_fuse,
                self.number_of_top_layers,
                self.fuse_continuously,
                self.remove_empty_layers,
            )
        return self._stacked_layers

    @property
    def at_layer(self):
        """StackedLayers for each cell."""
        return self.stacked_layers


################################################################################
# StackedLayers


class StackedLayers:
    """Track EventLayers where each event is its own layer and is made of
    multiple classes of material (e.g., different lithologies or grain sizes).

    Parameters
    ----------
    number_of_stacks : int
        Number of layer stacks to track.
    number_of_classes : int
        Number of material classes composing each layer.
    initial_allocation : int or array-like, optional
        Number of layers initially pre-allocated to speed up adding new layers.
        If an int, the layers are pre-allocated at the end; it an array-like,
        the first element corresponds to layers pre-allocated at the start, the
        second to layers pre-allocated at the end.
    new_allocation : int or array-like, optional
        Number of layers pre-allocated when adding new layers to speed up the
        process. If an array-like, different number of layers are pre-allocated
        at the start and at the end of the stacks.
    number_of_layers_to_fuse : int, optional
        Number of layers to merge together when calling `fuse`.
    number_of_top_layers : int, optional
        Number of layers at the top of the stack to ignore when calling `fuse`.
    fuse_continuously : bool, optional
        If True, fuse the layers as they are added; otherwise, wait until the
        number of unfused layers reaches `number_of_layers_to_fuse` (excluding
        the top layers) to fuse them.
    remove_empty_layers : bool, optional
        If True, remove the layers without any deposits at the top of the stack
        when adding a layer; otherwise, keep all the layers.
    """

    def __init__(
        self,
        number_of_stacks,
        number_of_classes=1,
        initial_allocation=0,
        new_allocation=1,
        number_of_layers_to_fuse=1,
        number_of_top_layers=1,
        fuse_continuously=True,
        remove_empty_layers=False,
    ):
        super().__init__()

        if isinstance(initial_allocation, int):
            initial_allocation = (0, initial_allocation)

        self._first_layer = initial_allocation[0]
        self._number_of_layers = 0
        self._number_of_stacks = number_of_stacks
        self._number_of_classes = number_of_classes
        self._surface_index = np.zeros(number_of_stacks, dtype=int)
        self._attrs = {}

        self.number_of_layers_to_fuse = number_of_layers_to_fuse
        self.number_of_top_layers = number_of_top_layers
        self.fuse_continuously = fuse_continuously
        self._number_of_fused_layers = 0
        self._number_of_sublayers = 0

        dims = (self.number_of_layers, self.number_of_stacks, self.number_of_classes)
        self._attrs["_dz"] = np.zeros(dims, dtype=float)
        self._resize(initial_allocation[0], initial_allocation[1])
        self._left_allocated = initial_allocation[0]
        self._right_allocated = initial_allocation[1]
        if isinstance(new_allocation, int):
            self.new_allocation = (new_allocation, new_allocation)
        else:
            self.new_allocation = new_allocation
        self.remove_empty_layers = remove_empty_layers

    def __getitem__(self, name):
        return self._attrs[name][
            self._first_layer : self._first_layer + self.number_of_layers
        ]

    def __setitem__(self, name, values):
        dims = (self.allocated, self.number_of_stacks)
        values = np.asarray(values)
        # Landlab's EventLayers expands and broadcasts the layers; it seems
        # unnecessary, but maybe it's more robust?
        # if values.ndim == 1:
        #     values = np.expand_dims(values, 1)
        # values = np.broadcast_to(values, shape)
        self._attrs[name] = _allocate_layers_for(values, *dims)
        self._attrs[name][
            self._first_layer : self._first_layer + self.number_of_layers
        ] = values

    def __iter__(self):
        return (name for name in self._attrs if not name.startswith("_"))

    def __str__(self):
        lines = [
            "number_of_layers: {number_of_layers}",
            "number_of_stacks: {number_of_stacks}",
            "number_of_classes: {number_of_classes}",
            "tracking: {attrs}",
        ]
        return os.linesep.join(lines).format(
            number_of_layers=self.number_of_layers,
            number_of_stacks=self.number_of_stacks,
            number_of_classes=self.number_of_classes,
            attrs=", ".join(self.tracking) or "null",
        )

    def __repr__(self):
        return (
            self.__class__.__name__
            + "({number_of_stacks}, {number_of_classes})".format(
                number_of_stacks=self.number_of_stacks,
                number_of_classes=self.number_of_classes,
            )
        )

    @property
    def tracking(self):
        """Layer properties being tracked."""
        return [name for name in self._attrs if not name.startswith("_")]

    def _setup_layers(self, **kwds):
        dims = (self.allocated, self.number_of_stacks)
        for name, array in kwds.items():
            self._attrs[name] = _allocate_layers_for(array, *dims)

    @property
    def number_of_stacks(self):
        """Number of stacks."""
        return self._number_of_stacks

    @property
    def number_of_classes(self):
        """Number of classes."""
        return self._number_of_classes

    @property
    def thickness(self):
        """Total thickness of the columns.

        The sum of all layer thicknesses for each stack as an array
        of shape `(number_of_stacks, )`.
        """
        return np.sum(self.dz, axis=(0, 2))

    def _get_thickness(self, axis, porosity=None):
        """Total sediment thickness of the columns, removing the porosity if given."""
        if porosity is None:
            return np.sum(self.dz, axis=axis)
        elif isinstance(porosity, str) and porosity in self._attrs:
            return np.sum(self.dz * (1.0 - self[porosity]), axis=axis)
        else:
            porosity = np.asarray(porosity)
            if porosity.shape == self.dz.shape[-len(porosity.shape) :]:
                return np.sum(self.dz * (1.0 - porosity), axis=axis)
            else:
                return np.sum(self.dz, axis=axis) * (1.0 - porosity)

    def get_thickness(self, porosity=None):
        """Total sediment thickness of the columns, removing the porosity if given.

        Parameters
        ----------
        porosity : array_like or str
            Porosity of all the layers for each class, which can be a layer
            property given by its name, or the porosity for each class, which
            is then identical for all layers.

        Returns
        -------
        thickness
            The sum of all layer thicknesses for each stack as an array of shape
            `(number_of_stacks,)`.
        """
        return self._get_thickness((0, 2), porosity=porosity)

    @property
    def class_thickness(self):
        """Total thickness of the columns for each class.

        The sum of all layer thicknesses for each stack as an array
        of shape `(number_of_stacks, number_of_classes)`.
        """
        return np.sum(self.dz, axis=0)

    def get_class_thickness(self, porosity=None):
        """Total sediment thickness of the columns for each class, removing the
        porosity if given.

        Parameters
        ----------
        porosity : array_like or str
            Porosity of all the layers for each class, which can be a layer
            property given by its name, or the porosity for each class, which
            is then identical for all layers.

        Returns
        -------
        thickness
            The sum of all layer thicknesses for each stack as an array of shape
            `(number_of_stacks, number_of_classes)`.
        """
        return self._get_thickness(0, porosity=porosity)

    @property
    def layer_thickness(self):
        """Thickness of each layer.

        The sum of all class thicknesses for each layer and stack as an array
        of shape `(number_of_layers, number_of_stacks)`.
        """
        return np.sum(self.dz, axis=2)

    @property
    def z(self):
        """Thickness to bottom of each layer.

        Thickness from the top of each stack to the bottom of each layer
        as an array of shape `(number_of_layers, number_of_stacks)`.
        """
        return np.cumsum(np.sum(self.dz[::-1], axis=2), axis=0)[::-1]

    @property
    def dz(self):
        """Thickness of each class in each layer.

        The thickness of each layer at each stack as an array of shape
        `(number_of_layers, number_of_stacks, number of classes)`.
        """
        return self._attrs["_dz"][
            self._first_layer : self._first_layer + self.number_of_layers
        ]

    def _get_composition(self, layer):
        """Get the composition of all classes in one or several layers."""
        thickness = np.sum(layer, axis=-1, keepdims=True)

        composition = np.zeros_like(layer)
        np.divide(layer, thickness, out=composition, where=thickness > 0.0)

        return composition

    @property
    def composition(self):
        """Composition of each layer.

        The composition of each layer at each stack as an array of shape
        `(number_of_layers, number_of_stacks, number of classes)`.
        """
        return self._get_composition(self.dz)

    @property
    def most_frequent_class(self):
        """The most frequent class of each layer.

        The most frequent class of each layer at each stack as an array of shape
        `(number_of_layers, number_of_stacks)`.
        """
        return np.argmax(self.composition, axis=2)

    @property
    def number_of_layers(self):
        """Total number of layers."""
        return self._number_of_layers

    @property
    def allocated(self):
        """Total number of allocated layers."""
        return self._attrs["_dz"].shape[0]

    def add(self, dz, at_bottom=False, update=False, update_compatible=False, **kwds):
        """Add a layer to the stack.

        Parameters
        ----------
        dz : float or array_like
            Thickness to add to each stack.
        at_bottom : bool, optional
            If False, add the layer to the top of the stack; otherwise, insert
            the layer at the bottom of the stack.
        update : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer at the top or bottom of the
            stack (depending on `at_bottom`).
        update_compatible : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer at the top or bottom of the
            stack (depending on `at_bottom`) only if the new layer is compatible
            with the existing layer.
        """
        if self.number_of_layers == 0:
            self._setup_layers(**kwds)

        if update_compatible == True:
            is_compatible = self.number_of_layers > 0 and self.is_compatible(dz, **kwds)
        else:
            is_compatible = False

        if (update == False and is_compatible == False) or (
            update == True and self.number_of_layers == 0
        ):
            self._add_empty_layer(at_bottom=at_bottom)

        layer = self._first_layer
        if at_bottom == False:
            layer += self.number_of_layers - 1
        _deposit_or_erode(self._attrs["_dz"], self._first_layer, layer, dz)
        last_layer = self._first_layer + self.number_of_layers - 1
        _get_surface_index(
            self._attrs["_dz"], self._first_layer, last_layer, self._surface_index
        )

        if is_compatible == False:
            for name in kwds:
                try:
                    self._attrs[name][layer] = kwds[name]
                except KeyError as exc:
                    raise ValueError(
                        f"{name!r} is not being tracked. Error in adding."
                    ) from exc

        if self.remove_empty_layers == True:
            self._remove_empty_layers()

    def update(self, layer, dz, **kwds):
        """Update a given layer in the stacks.

        Parameters
        ----------
        layer : int
            Index of the layer to update.
        dz : float or array_like
            Thickness to add to each stack.
        """
        if self.number_of_layers == 0:
            self._setup_layers(**kwds)

        _deposit_or_erode(
            self._attrs["_dz"], self._first_layer, self._first_layer + layer, dz
        )
        last_layer = self._first_layer + self.number_of_layers - 1
        _get_surface_index(
            self._attrs["_dz"], self._first_layer, last_layer, self._surface_index
        )

        for name in kwds:
            try:
                if kwds[name] is not None:
                    self._attrs[name][self._first_layer + layer] = kwds[name]
            except KeyError as exc:
                raise ValueError(
                    f"{name!r} is not being tracked. Error in adding."
                ) from exc

    def _reduce_attribute(self, array, start, stop, step, n_blocks, reducer):
        """Combines layers of a specific attribute."""
        if self.fuse_continuously == True and reducer == np.mean and stop - start == 2:
            factor = np.array([[self._number_of_sublayers], [1]])
            factor = factor.reshape(factor.shape + (1,) * (array.ndim - factor.ndim))
            middle = _reduce_matrix(factor * array[start:stop], step, np.sum) / (
                self._number_of_sublayers + 1.0
            )
        elif reducer == "weighted_mean":
            thickness = self._attrs["_dz"][start:stop]
            total_thickness = np.sum(thickness, axis=0, keepdims=True)
            factor = np.zeros_like(thickness)
            np.divide(
                thickness, total_thickness, out=factor, where=total_thickness > 0.0
            )
            middle = _reduce_matrix(factor * array[start:stop], step, np.sum)
        else:
            middle = _reduce_matrix(array[start:stop], step, reducer)
        top = array[stop : self._first_layer + self._number_of_layers]

        array[start : start + n_blocks] = middle
        array[start + n_blocks : start + n_blocks + len(top)] = top

    def reduce(self, *args, **kwds):
        """reduce([start], stop, [step])
        Combine layers.

        Reduce adjacent layers into a single layer. Layer reduction can be done
        using a function (e.g., np.sum or np.mean) or with the string 'weighted_mean',
        which uses a mean weighted by layer thickness.
        """
        _valid_keywords_or_raise(kwds, required=self.tracking, optional=self._attrs)

        start, stop, step = _BlockSlice(*args).indices(self._number_of_layers)
        start += self._first_layer
        stop += self._first_layer

        if step <= 1:
            return

        n_blocks = (stop - start) // step
        n_removed = n_blocks * (step - 1)
        for name, array in self._attrs.items():
            if name != "_dz":
                reducer = kwds.get(name, np.sum)
                self._reduce_attribute(array, start, stop, step, n_blocks, reducer)
        reducer = kwds.get("_dz", np.sum)
        self._reduce_attribute(self._attrs["_dz"], start, stop, step, n_blocks, reducer)

        self._number_of_layers -= n_removed
        self._right_allocated = (
            len(self._attrs["_dz"]) - self._first_layer - self._number_of_layers
        )
        self._left_allocated = self._first_layer
        last_layer = self._first_layer + self.number_of_layers - 1
        _get_surface_index(
            self._attrs["_dz"], self._first_layer, last_layer, self._surface_index
        )

    def fuse(self, finalize=False, **kwds):
        """Fuse layers to save computational resources during runtime. Layers
        are fused together from the top based on `number_of_layers_to_fuse` and,
        once fused, are not fused again. The top layers defined by
        `number_of_top_layers` are not fused unless `finalize` is True.

        Parameters
        ----------
        finalize : bool, optional
            If True, all the layers are fused, including the top layers defined
            by `number_of_top_layers`; otherwise, the top layers are ignored.
        """
        start_fuse = self._number_of_fused_layers
        stop_fuse = self._number_of_layers

        if finalize == False:
            stop_fuse -= self.number_of_top_layers
            if (
                self.fuse_continuously == False
                and stop_fuse - start_fuse >= self.number_of_layers_to_fuse
            ):
                self.reduce(
                    start_fuse, stop_fuse, self.number_of_layers_to_fuse, **kwds
                )
                self._number_of_fused_layers += (
                    stop_fuse - start_fuse
                ) // self.number_of_layers_to_fuse
            elif self.fuse_continuously == True and stop_fuse > start_fuse:
                self.reduce(start_fuse, stop_fuse, **kwds)
                # Increasing the number of sublayers here means that fuse needs
                # to be called multiple times when multiple components are used
                # to get the proper behavior
                self._number_of_sublayers += 1
                if self._number_of_sublayers == self.number_of_layers_to_fuse:
                    self._number_of_sublayers = 0
                    self._number_of_fused_layers += 1
        else:
            if stop_fuse - start_fuse > self.number_of_layers_to_fuse:
                self.reduce(
                    start_fuse, stop_fuse, self.number_of_layers_to_fuse, **kwds
                )
                self._number_of_fused_layers += (
                    stop_fuse - start_fuse
                ) // self.number_of_layers_to_fuse
                start_fuse = self._number_of_fused_layers
                stop_fuse = self._number_of_layers
            self.reduce(start_fuse, stop_fuse, **kwds)
            self._number_of_fused_layers += 1

    @property
    def surface_index(self):
        """Index to the top non-empty layer."""
        return self._surface_index

    def get_surface_values(self, name):
        """Values of a field on the surface layer."""
        return self._attrs[name][self.surface_index, np.arange(self._number_of_stacks)]

    def get_surface_composition(self):
        """Composition of the surface layer (i.e., proportion of each class)."""
        return self._get_composition(self.get_surface_values("_dz"))

    def get_superficial_layer(self, dz):
        """Get the thicknesses of all classes in a superficial layer defined by
        its thickness from the surface `dz`.

        Parameters
        ----------
        dz : float or array_like
            Thickness from the surface of the superficial layer.

        Returns
        -------
        superficial_layer
            The thickness of material from each class within the superficial layer.
        """
        superficial_layer = np.zeros((self.number_of_stacks, self.number_of_classes))
        _get_superficial_layer(
            self._attrs["_dz"],
            self._first_layer,
            self._surface_index,
            dz,
            superficial_layer,
        )

        return superficial_layer

    def get_superficial_composition(self, dz):
        """Get the composition of all classes in a superficial layer defined by
        its thickness from the surface `dz`.

        Parameters
        ----------
        dz : float or array_like
            Thickness from the surface of the superficial layer.

        Returns
        -------
        superficial_composition
            The composition of material from each class within the superficial layer.
        """
        return self._get_composition(self.get_superficial_layer(dz))

    def get_active_layer(self, dz, porosity=None):
        """Get the thicknesses of all classes in an active layer defined by its
        thickness from the surface `dz`.

        Parameters
        ----------
        dz : float or array_like
            Thickness from the surface of the active layer.
        porosity : array_like or str
            Porosity of all the layers for each class, which can be a layer
            property given by its name, or the porosity for each class, which
            is then identical for all layers.

        Returns
        -------
        active_layer
            The thickness of material from each class within the active layer.
        """
        if isinstance(porosity, str):
            porosity = self._attrs[porosity]
        elif porosity is not None:
            porosity = np.asarray(porosity)

        active_layer = np.zeros((self.number_of_stacks, self.number_of_classes))
        _get_active_layer(
            self._attrs["_dz"],
            porosity,
            self._first_layer,
            self._surface_index,
            dz,
            active_layer,
        )

        return active_layer

    def get_active_composition(self, dz, porosity=None):
        """Get the composition of all classes in an active layer defined by its
        thickness from the surface `dz`.

        Parameters
        ----------
        dz : float or array_like
            Thickness from the surface of the active layer.
        porosity : array_like or str
            Porosity of all the layers for each class, which can be a layer
            property given by its name, or the porosity for each class, which
            is then identical for all layers.

        Returns
        -------
        active_composition
            The composition of material from each class within the active layer.
        """
        return self._get_composition(self.get_active_layer(dz, porosity))

    def _add_empty_layer(self, at_bottom=False):
        """Add a new empty layer to the stacks."""
        if at_bottom == False and self._right_allocated == 0:
            self._resize(0, self.new_allocation[1])
            self._right_allocated = self.new_allocation[1]
        elif at_bottom == True and self._left_allocated == 0:
            self._resize(self.new_allocation[0], 0)
            self._left_allocated = self.new_allocation[0]
            self._first_layer = self.new_allocation[0]

        self._number_of_layers += 1
        if at_bottom == False:
            self._right_allocated -= 1
            layer = self._first_layer + self.number_of_layers - 1
        else:
            self._left_allocated -= 1
            self._first_layer -= 1
            layer = self._first_layer
        self._attrs["_dz"][layer, :] = 0.0
        for name in self._attrs:
            self._attrs[name][layer] = 0.0

    def _remove_empty_layers(self):
        """Remove empty layers at the top of the stack"""
        number_of_filled_layers = self.surface_index.max() + 1 - self._first_layer
        if number_of_filled_layers < self.number_of_layers:
            self._number_of_layers = number_of_filled_layers

    def remove_last_layers(self, number_of_layers=1):
        """Remove the last layers at the top of the stack

        Parameters
        ----------
        number_of_layers : int
            The number of layers to remove
        """
        self._number_of_layers -= number_of_layers
        self._right_allocated += number_of_layers
        self._surface_index -= number_of_layers

    def is_compatible(self, dz, **kwds):
        """Check if a new layer is compatible with the existing top layer.

        Parameters
        ----------
        dz : float or array_like
            Thickness to add to each stack.

        Returns
        -------
        bool
            ``True`` if the new layer is compatible, otherwise ``False``.
        """
        if isinstance(dz, np.ndarray) == True and dz.ndim == 2:
            dz = np.sum(dz, axis=1)

        where_deposition = dz > 0.0

        if np.any(where_deposition):
            if not_tracked := set(kwds) - set(self):
                raise ValueError(
                    "Error adding layer."
                    f" {', '.join(sorted(repr(t) for t in not_tracked))}"
                    " is not being tracked. Currently tracking:"
                    f" {', '.join(sorted(repr(t) for t in set(self)))}"
                )
            for name in kwds:
                is_compatible = self[name][-1] == kwds[name]

                if not np.all(is_compatible[where_deposition]):
                    return False
        return True

    def _resize(self, left_new_cap, right_new_cap):
        """Allocate more memory for the layers."""
        for name in self._attrs:
            self._attrs[name] = resize_array(
                self._attrs[name], left_new_cap, right_new_cap
            )



"""Cyclic sea level calculator"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np
from landlab import Component

from ...utils import convert_to_array, match_array_lengths


################################################################################
# Component


class CyclicSeaLevelCalculator(Component):
    """Calculate sea level based on a mixture of sinusoids."""

    _name = "CyclicSeaLevelCalculator"

    _unit_agnostic = True

    _info = {
        "topographic__elevation": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "sea_level__elevation": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "m",
            "mapping": "grid",
            "doc": "Sea level elevation",
        },
        "bathymetric__depth": {
            "dtype": float,
            "intent": "out",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The water depth under the sea",
        },
    }

    def __init__(
        self,
        grid,
        wavelength=100000.0,
        time_shift=0.0,
        amplitude=5.0,
        mean=0.0,
        rate=0.0,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        wavelength : float or array
            The wavelength(s) of the sinusoid(s) driving sea level variation.
        time_shift : float or array
            The time shift(s) of the sinusoid(s) driving sea level variation,
            which control(s) sea level at time 0.
        amplitude : float or array
            The amplitude(s) of the sinusoid(s) driving sea level variation.
        mean : float
            The mean sea level.
        rate : float
            The linear rate of sea level increase.
        """
        super().__init__(grid)

        self._grid.at_grid["sea_level__elevation"] = 0.
        self.initialize_output_fields()

        # Parameters
        self.wavelength = convert_to_array(wavelength)
        self.time_shift = convert_to_array(time_shift)
        self.amplitude = convert_to_array(amplitude)
        self.wavelength, self.time_shift, self.amplitude = match_array_lengths(
            self.wavelength, self.time_shift, self.amplitude
        )
        self.mean = mean
        self.rate = rate
        self._time = 0.0

        # Physical fields
        self._topography = grid.at_node["topographic__elevation"]
        self._bathymetry = grid.at_node["bathymetric__depth"]

        # Initialize the fields
        self.run_one_step(0.0)

    def run_one_step(self, dt):
        """Run the calculator for one timestep, dt.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        """
        self._time += dt

        sea_level = self.mean + self.rate * self._time
        for wavelength, time_shift, amplitude in zip(
            self.wavelength, self.time_shift, self.amplitude
        ):
            sea_level += amplitude * np.sin(
                2.0 * np.pi * (self._time - time_shift) / wavelength
            )
        self._grid.at_grid["sea_level__elevation"] = sea_level

        np.subtract(
            sea_level,
            self._topography,
            out=self._bathymetry,
            where=(
                self._topography < sea_level
            ),  # & (self._grid.status_at_node == self._grid.BC_NODE_IS_CORE),
        )



"""Cython functions for diffusion"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


cimport cython

# https://cython.readthedocs.io/en/stable/src/userguide/fusedtypes.html
ctypedef fused id_t:
    cython.integral
    long long


################################################################################
# Functions


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void calculate_sediment_fluxes(
    const id_t [:] node_order,
    const id_t [:, :] neighbors,
    const cython.floating [:, :] flux_proportions,
    cython.floating [:, :] sediment_influx,
    cython.floating [:, :, :] sediment_outflux,
    const cython.floating [:, :, :] transferred_fraction,
    const cython.floating [:, :, :] erosion_flux_sed,
    const cython.floating [:, :] max_sediment_outflux,
) noexcept nogil:
    """Calculates sediment influx."""
    cdef unsigned int n_nodes = node_order.shape[0]
    cdef unsigned int n_receivers = neighbors.shape[1]
    cdef unsigned int n_sediments = sediment_outflux.shape[2]
    cdef unsigned int node, i, j, k
    cdef double total_outflux, max_outflux
    cdef double limiter

    # Iterate top to bottom through the nodes, update sediment out- and influx.
    # Because calculation of the outflux requires the influx, this operation
    # must be done in an upstream to downstream loop, and cannot be vectorized.
    for i in range(n_nodes - 1, -1, -1):

        # Choose the node id
        node = node_order[i]

        # For each sediment class...
        for k in range(n_sediments):

            # Compute the available sediments, i.e., the maximum ouflux
            max_outflux = sediment_influx[node, k] + max_sediment_outflux[node, k]
            if max_outflux > 0.:
                # Compute the sediment outflux
                for j in range(n_receivers):
                    sediment_outflux[node, j, k] = erosion_flux_sed[node, j, k]
                    sediment_outflux[node, j, k] += flux_proportions[node, j]*transferred_fraction[node, j, k]*sediment_influx[node, k]
                # Compute the total outflux
                total_outflux = 0.
                for j in range(n_receivers):
                    total_outflux += sediment_outflux[node, j, k]
                if total_outflux > 0.:
                    # Determine by how much the sediment outflux needs to be decreased
                    limiter = max_outflux/total_outflux
                    if limiter < 1.:
                        # Update the sediment outflux
                        for j in range(n_receivers):
                            sediment_outflux[node, j, k] *= limiter

                    # Add the outflux to the influx of the downstream node(s)
                    for j in range(n_receivers):
                        # TODO: Check this, it's not in the Landlab components, but it seems that it can fail otherwise
                        if neighbors[node, j] > -1:
                            sediment_influx[neighbors[node, j], k] += sediment_outflux[node, j, k]



"""Base components"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np
from landlab import Component

from ..utils import convert_to_array, format_fields_to_track


################################################################################
# Base components


class _BaseHandler(Component):
    """Base class to handle sediments in stacked layers."""

    _name = "_BaseHandler"

    _unit_agnostic = True

    def __init__(
        self,
        grid,
        fields_to_track=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(grid)

        # Parameters
        self._fields_to_track = format_fields_to_track(fields_to_track)

        # Physical fields
        self._stratigraphy = grid.stacked_layers
        self._time = 0.0

        # Fields for sediment fluxes
        self._sediment_thickness = np.zeros(
            (grid.number_of_nodes, self._stratigraphy.number_of_classes)
        )

    def _update_stratigraphy(
        self, dt, update_compatible=False, update=False, at_bottom=False
    ):
        """
        Updates the stratigraphy based on the sediment changes.
        """
        core_nodes = self._grid.core_nodes

        self._time += dt

        fields_to_track = {
            field: self._grid.at_node[field][core_nodes]
            for field in self._fields_to_track
        }
        # _sediment_thickness also includes bedrock erosion, which only affects
        # the topography (sediment thickness in _stratigraphy won't become negative
        # if more material is to be removed than what is available, and the
        # bedrock below that is considered infinite)
        self._stratigraphy.add(
            self._sediment_thickness[core_nodes],
            at_bottom=at_bottom,
            update=update,
            update_compatible=update_compatible,
            time=self._time,
            **fields_to_track,
        )


class _BaseMover(_BaseHandler):
    """Base class to move sediments in continental and marine domains."""

    _name = "_BaseMover"

    _unit_agnostic = True

    _info = {
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    def __init__(
        self,
        grid,
        fields_to_track=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(grid=grid, fields_to_track=fields_to_track, **kwargs)

        # Physical fields
        self._topography = grid.at_node["topographic__elevation"]

    def _update_physical_fields(self, dt, update_compatible=False, update=False):
        """
        Applies the sediment fluxes to the topography and stratigraphy.
        """
        core_nodes = self._grid.core_nodes

        self._update_stratigraphy(dt, update_compatible, update)
        self._topography[core_nodes] += np.sum(
            self._sediment_thickness[core_nodes], axis=1
        )


################################################################################
# Base components for diffusion models


class _BaseDiffuser(_BaseMover):
    """Base class to diffuse sediments in continental and marine domains."""

    _name = "_BaseDiffuser"

    _unit_agnostic = True

    _info = {
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "bathymetric__depth": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The water depth under the sea",
        },
    }

    def __init__(
        self,
        grid,
        diffusivity_cont=0.01,
        diffusivity_mar=0.001,
        wave_base=20.0,
        porosity=0.0,
        max_erosion_rate_sed=0.01,
        active_layer_rate_sed=None,
        bedrock_composition=1.0,
        max_erosion_rate_br=0.01,
        active_layer_rate_br=None,
        exponent_slope=1.0,
        fields_to_track=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        diffusivity_cont : float or array-like (m/time)
            The diffusivity of the sediments over the continental domain for one
            or multiple lithologies.
        diffusivity_mar : float or array-like (m/time)
            The diffusivity of the sediments over the marine domain for one or
            multiple lithologies.
        wave_base : float (m)
            The wave base, below which weathering decreases exponentially.
        porosity : float or array-like (-)
            The porosity of the sediments at the time of deposition for one or
            multiple lithologies. When computing the active layer, this porosity
            is used unless the field 'sediment__porosity' is being tracked in
            the stratigraphy.
        max_erosion_rate_sed : float (m/time), optional
            The maximum erosion rate of the sediments. If None, all the sediments
            may be eroded in a single time step. The erosion rate defines the
            thickness of the active layer of the sediments if `active_layer_rate`
            is None.
        active_layer_rate_sed : float (m/time), optional
            The rate of formation of the active layer for sediments, which is used
            to determine the composition of the transported sediments. By default,
            it is set by the maximum erosion rate of the sediments.
        bedrock_composition : float or array-like (-)
            The composition of the material is added to the StackedLayers from
            the bedrock.
        max_erosion_rate_br : float (m/time)
            The maximum erosion rate of the bedrock. The erosion rate defines the
            thickness of the active layer of the bedrock if `active_layer_rate`
            is None.
        active_layer_rate_br : float (m/time), optional
            The rate of formation of the active layer for the bedrock, which is
            used to determine the composition of the transported sediments. By
            default, it is set by the maximum erosion rate of the bedrock.
        exponent_slope : float (-)
            The exponent for the slope.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(grid=grid, fields_to_track=fields_to_track, **kwargs)

        # Parameters
        self._K_cont = convert_to_array(diffusivity_cont)
        self._K_mar = convert_to_array(diffusivity_mar)
        self.wave_base = wave_base
        self._porosity = convert_to_array(porosity)
        self._max_erosion_rate_sed = (
            np.inf if max_erosion_rate_sed is None else max_erosion_rate_sed
        )
        self._active_layer_rate_sed = (
            self._max_erosion_rate_sed
            if active_layer_rate_sed is None
            else active_layer_rate_sed
        )
        self._bedrock_composition = convert_to_array(bedrock_composition)
        self.max_erosion_rate_br = max_erosion_rate_br
        self._active_layer_rate_br = (
            self.max_erosion_rate_br
            if active_layer_rate_br is None
            else active_layer_rate_br
        )
        self._n = exponent_slope

        # Physical fields
        self._bathymetry = grid.at_node["bathymetric__depth"][:, np.newaxis]

        # Fields for sediment fluxes
        n_nodes = grid.number_of_nodes
        n_sediments = self._stratigraphy.number_of_classes
        self._K_sed = np.zeros((n_nodes, 1, n_sediments))
        self._active_layer = np.zeros((n_nodes, 1, n_sediments))
        self._active_layer_thickness = np.zeros((n_nodes, 1, 1))
        self._active_layer_composition = np.zeros((n_nodes, 1, n_sediments))
        self._total_layer_thickness = np.zeros((n_nodes, 1, 1))

    def _calculate_sediment_diffusivity(self):
        """
        Calculates the diffusivity coefficient of the sediments over the continental
        and marine domains.
        """
        self._K_sed[self._bathymetry[:, 0] == 0.0] = self._K_cont
        self._K_sed[self._bathymetry[:, 0] > 0.0, 0] = self._K_mar * np.exp(
            -self._bathymetry[self._bathymetry[:, 0] > 0.0] / self.wave_base
        )

    def _calculate_active_layer(self, max_thickness_sed, max_thickness_br):
        """
        Calculates the active layer based on the sediments and the bedrock.
        """
        porosity = (
            "sediment__porosity"
            if "sediment__porosity" in self._stratigraphy._attrs
            else self._porosity
        )
        self._active_layer[self._grid.core_nodes, 0] = (
            self._stratigraphy.get_active_layer(max_thickness_sed, porosity)
        )
        if max_thickness_br > 0.0:
            self._active_layer_thickness[:] = np.sum(
                self._active_layer, axis=2, keepdims=True
            )
            # TODO: This seems to really slow things down
            self._total_layer_thickness[self._grid.core_nodes, 0, 0] = (
                self._stratigraphy.get_thickness(porosity)
            )
            np.add(
                self._active_layer,
                self._bedrock_composition
                * (
                    (1.0 - self._porosity) * max_thickness_br
                    - self._active_layer_thickness
                ),
                out=self._active_layer,
                where=(
                    np.abs(self._active_layer_thickness - self._total_layer_thickness)
                    < 1e-10
                )
                & (
                    self._active_layer_thickness
                    < (1.0 - self._porosity) * max_thickness_br
                ),
            )

    def _calculate_active_layer_composition(self, dt):
        """
        Calculates the composition of the active layer based on the sediments
        and the bedrock.
        """
        self._calculate_active_layer(
            self._active_layer_rate_sed * dt, self._active_layer_rate_br * dt
        )

        self._active_layer_thickness[:] = np.sum(
            self._active_layer, axis=2, keepdims=True
        )
        self._active_layer_composition[:] = 0.0
        np.divide(
            self._active_layer,
            self._active_layer_thickness,
            out=self._active_layer_composition,
            where=self._active_layer_thickness > 0.0,
        )


class _BaseStreamPower(_BaseDiffuser):
    """Base class to move sediments in continental and marine domains using the
    stream power law.
    """

    _name = "_BaseStreamPower"

    _unit_agnostic = True

    _info = {
        "flow__upstream_node_order": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array containing downstream-to-upstream ordered list of node IDs",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "flow__receiver_proportions": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of proportion of flow sent to each receiver.",
        },
        "water__unit_flux_in": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m/s",
            "mapping": "node",
            "doc": "External volume water per area per time input to each node (e.g., rainfall rate)",
        },
        "sediment__unit_flux_in": {
            "dtype": float,
            "intent": "in",
            "optional": True,
            "units": "m3/s",
            "mapping": "node",
            "doc": "Sediment flux as boundary condition",
        },
        "surface_water__discharge": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "m3/s",
            "mapping": "node",
            "doc": "Volumetric discharge of surface water",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
        "bathymetric__depth": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The water depth under the sea",
        },
    }

    def __init__(
        self,
        grid,
        diffusivity_cont=0.01,
        diffusivity_mar=0.001,
        wave_base=20.0,
        critical_flux=0.0,
        porosity=0.0,
        max_erosion_rate_sed=0.01,
        active_layer_rate_sed=None,
        bedrock_composition=1.0,
        max_erosion_rate_br=0.01,
        active_layer_rate_br=None,
        exponent_discharge=1.0,
        exponent_slope=1.0,
        ref_water_flux=None,
        fields_to_track=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        diffusivity_cont : float or array-like (m/time)
            The diffusivity of the sediments over the continental domain for one
            or multiple lithologies.
        diffusivity_mar : float or array-like (m/time)
            The diffusivity of the sediments over the marine domain for one or
            multiple lithologies.
        wave_base : float (m)
            The wave base, below which weathering decreases exponentially.
        critical_flux : float or array-like (m3/time)
            Critical sediment flux to start displace sediments in the stream
            power law.
        porosity : float or array-like (-)
            The porosity of the sediments at the time of deposition for one or
            multiple lithologies. When computing the active layer, this porosity
            is used unless the field 'sediment__porosity' is being tracked in
            the stratigraphy.
        max_erosion_rate_sed : float (m/time), optional
            The maximum erosion rate of the sediments. If None, all the sediments
            may be eroded in a single time step. The erosion rate defines the
            thickness of the active layer of the sediments if `active_layer_rate`
            is None.
        max_erosion_rate_br : float (m/time)
            The maximum erosion rate of the bedrock. The erosion rate defines the
            thickness of the active layer of the bedrock if `active_layer_rate`
            is None.
        active_layer_rate_sed : float (m/time), optional
            The rate of formation of the active layer for sediments, which is used
            to determine the composition of the transported sediments. By default,
            it is set by the maximum erosion rate of the sediments.
        active_layer_rate_br : float (m/time), optional
            The rate of formation of the active layer for the bedrock, which is
            used to determine the composition of the transported sediments. By
            default, it is set by the maximum erosion rate of the bedrock.
        bedrock_composition : float or array-like (-)
            The composition of the material is added to the StackedLayers from
            the bedrock.
        exponent_discharge : float (-)
            The exponent for the water discharge.
        exponent_slope : float (-)
            The exponent for the slope.
        ref_water_flux : float or string (m3/time), optional
            The reference water flux by which the water discharge is normalized.
            If a float, that value is used at each time step; if 'max', the
            maximum value of discharge at each time step is used.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(
            grid,
            diffusivity_cont,
            diffusivity_mar,
            wave_base,
            porosity,
            max_erosion_rate_sed,
            active_layer_rate_sed,
            bedrock_composition,
            max_erosion_rate_br,
            active_layer_rate_br,
            exponent_slope,
            fields_to_track,
        )

        # Parameters
        self._critical_flux = convert_to_array(critical_flux)
        self._m = exponent_discharge
        self.ref_water_flux = ref_water_flux

        # Fields for stream power
        self._node_order = grid.at_node["flow__upstream_node_order"]
        self._flow_receivers = grid.at_node["flow__receiver_node"][..., np.newaxis]
        self._link_to_receiver = grid.at_node["flow__link_to_receiver_node"][
            ..., np.newaxis
        ]
        self._slope = grid.at_node["topographic__steepest_slope"][..., np.newaxis]
        if self._flow_receivers.ndim == 2:
            self._flow_receivers = self._flow_receivers[..., np.newaxis]
            self._link_to_receiver = self._link_to_receiver[..., np.newaxis]
            self._slope = self._slope[..., np.newaxis]
        n_nodes = grid.number_of_nodes
        n_receivers = self._flow_receivers.shape[1]
        if "flow__receiver_proportions" in grid.at_node:
            self._flow_proportions = grid.at_node["flow__receiver_proportions"][
                ..., np.newaxis
            ]
        else:
            self._flow_proportions = np.ones((n_nodes, n_receivers, 1))
        self._water_flux = grid.at_node["surface_water__discharge"][
            :, np.newaxis, np.newaxis
        ]

        # Fields for sediment fluxes
        n_sediments = self._stratigraphy.number_of_classes
        if "sediment__unit_flux_in" in grid.at_node:
            self._sediment_input = grid.at_node["sediment__unit_flux_in"]
            if self._sediment_input.ndim == 1:
                self._sediment_input = self._sediment_input[:, np.newaxis]
        else:
            self._sediment_input = np.zeros((n_nodes, n_sediments))
        self._critical_rate = np.zeros((n_nodes, 1, n_sediments))
        self._ratio_critical_outflux = np.zeros((n_nodes, n_receivers, n_sediments))

    def _normalize_water_flux(self):
        """
        Normalizes the water flux if needed.
        """
        if self.ref_water_flux == "max":
            self._water_flux[:] /= np.max(self._water_flux)
        elif isinstance(self.ref_water_flux, (int, float)):
            self._water_flux[:] /= self.ref_water_flux


################################################################################
# Base components for routing models


class _BaseRouter(_BaseMover):
    """Base class to route sediments in continental and marine domains."""

    _name = "_BaseRouter"

    _unit_agnostic = True

    _info = {
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
    }

    def __init__(
        self,
        grid,
        number_of_neighbors=1,
        porosity=0.0,
        fields_to_track=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        number_of_neighbors : int, optional
            The number of neighbors around a cell to consider when computing
            sediment displacement.
        porosity : float or array-like (-)
            The porosity of the sediments at the time of deposition for one or
            multiple lithologies.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(grid=grid, fields_to_track=fields_to_track, **kwargs)

        # Parameters
        self._porosity = convert_to_array(porosity)

        # Fields for sediment fluxes
        n_nodes = grid.number_of_nodes
        n_sediments = self._stratigraphy.number_of_classes
        self._sediment_influx = np.zeros((n_nodes, n_sediments))
        self._sediment_outflux = np.zeros((n_nodes, number_of_neighbors, n_sediments))
        self._max_sediment_outflux = np.zeros((n_nodes, n_sediments))

    def _apply_fluxes(self, dt, update_compatible=False, update=False):
        """
        Applies the sediment fluxes to the topography and stratigraphy.
        """
        core_nodes = self._grid.core_nodes
        cell_area = self._grid.cell_area_at_node[:, np.newaxis]

        self._sediment_thickness[core_nodes] = (
            (
                self._sediment_influx[core_nodes]
                - np.sum(self._sediment_outflux[core_nodes], axis=1)
            )
            * dt
            / (1.0 - self._porosity)
            / cell_area[core_nodes]
        )
        self._update_physical_fields(dt, update_compatible, update)



from .grids import RasterModelGrid
from .grids import VoronoiDelaunayGrid
from .grids import FramedVoronoiGrid
from .grids import HexModelGrid

__all__ = [
    "HexModelGrid",
    "RasterModelGrid",
    "FramedVoronoiGrid",
    "VoronoiDelaunayGrid",
]



"""Cython functions for layers"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


cimport cython

# https://cython.readthedocs.io/en/stable/src/userguide/fusedtypes.html
ctypedef fused id_t:
    cython.integral
    long long


################################################################################
# Functions


@cython.boundscheck(False)
@cython.wraparound(False)
def deposit_or_erode(
    cython.floating [:, :, :] layers,
    long bottom_index,
    long top_index,
    const cython.floating [:, :] dz,
):
    cdef int n_stacks = layers.shape[1]
    cdef int n_classes = layers.shape[2]
    cdef int col
    cdef int cla
    cdef int layer
    cdef double removed
    cdef double amount_to_remove

    with nogil:
        for col in range(n_stacks):
            for cla in range(n_classes):
                if dz[col, cla] >= 0.:
                    layers[top_index, col, cla] += dz[col, cla]
                else:
                    amount_to_remove = - dz[col, cla]
                    removed = 0.
                    for layer in range(top_index, bottom_index - 1, -1):
                        removed += layers[layer, col, cla]
                        layers[layer, col, cla] = 0.
                        if removed > amount_to_remove:
                            layers[layer, col, cla] = removed - amount_to_remove
                            break


@cython.boundscheck(False)
@cython.wraparound(False)
def get_surface_index(
    const cython.floating [:, :, :] layers,
    long bottom_index,
    long top_index,
    id_t [:] surface_index,
):
    cdef int n_stacks = layers.shape[1]
    cdef int n_classes = layers.shape[2]
    cdef int col
    cdef int cla
    cdef int layer

    with nogil:
        for col in range(n_stacks):
            surface_index[col] = bottom_index
            for layer in range(top_index, bottom_index - 1, -1):
                for cla in range(n_classes):
                    if layers[layer, col, cla] > 0:
                        surface_index[col] = layer
                        break
                else:
                    continue
                break


@cython.boundscheck(False)
@cython.wraparound(False)
def get_superficial_layer(
    const cython.floating [:, :, :] layers,
    long bottom_index,
    const id_t [:] surface_index,
    const cython.floating [:] dz,
    cython.floating [:, :] superficial_layer,
):
    cdef int n_stacks = layers.shape[1]
    cdef int n_classes = layers.shape[2]
    cdef int col
    cdef int cla
    cdef int layer
    cdef double thickness
    cdef double superficial_layer_thickness
    cdef double layer_thickness
    cdef double ratio

    with nogil:
        for col in range(n_stacks):
            superficial_layer_thickness = dz[col]
            thickness = 0.
            for layer in range(surface_index[col], bottom_index - 1, -1):
                for cla in range(n_classes):
                    superficial_layer[col, cla] += layers[layer, col, cla]
                    thickness += layers[layer, col, cla]
                if thickness > superficial_layer_thickness:
                    layer_thickness = 0.
                    for cla in range(n_classes):
                        layer_thickness += layers[layer, col, cla]
                    ratio = (thickness - superficial_layer_thickness)/layer_thickness
                    for cla in range(n_classes):
                        superficial_layer[col, cla] -= ratio*layers[layer, col, cla]
                    break


@cython.boundscheck(False)
@cython.wraparound(False)
def get_active_layer(
    const cython.floating [:, :, :] layers,
    const cython.floating [:, :, :] porosity,
    long bottom_index,
    const id_t [:] surface_index,
    const cython.floating [:] dz,
    cython.floating [:, :] active_layer,
):
    cdef int n_stacks = layers.shape[1]
    cdef int n_classes = layers.shape[2]
    cdef int col
    cdef int cla
    cdef int layer
    cdef double thickness
    cdef double active_layer_thickness
    cdef double layer_thickness
    cdef double ratio

    with nogil:
        for col in range(n_stacks):
            active_layer_thickness = dz[col]
            thickness = 0.
            for layer in range(surface_index[col], bottom_index - 1, -1):
                for cla in range(n_classes):
                    active_layer[col, cla] += (1. - porosity[layer, col, cla])*layers[layer, col, cla]
                    thickness += layers[layer, col, cla]
                if thickness > active_layer_thickness:
                    layer_thickness = 0.
                    for cla in range(n_classes):
                        layer_thickness += layers[layer, col, cla]
                    ratio = (thickness - active_layer_thickness)/layer_thickness
                    for cla in range(n_classes):
                        active_layer[col, cla] -= ratio*(1. - porosity[layer, col, cla])*layers[layer, col, cla]
                    break



"""Simple sediment landslider"""

# MIT License

# Copyright (c) 2025 Guillaume Rongier

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import numpy as np
from landlab import RasterModelGrid

from .._base import _BaseRouter
from ...utils import convert_to_array
from .cfuncs import calculate_sediment_influx


################################################################################
# Component


class SimpleSedimentLandslider(_BaseRouter):
    """Simple slope failure and mass flows of sediments in a StackedLayers.

    References
    ----------
    Granjeon, D. (2014)
        3D forward modelling of the impact of sediment transport and base level cycles on continental margins and incised valleys
        https://doi.org/10.1002/9781118920435.ch16
    """

    _name = "SimpleSedimentLandslider"

    _unit_agnostic = True

    _info = {
        "flow__link_to_receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "ID of link downstream of each node, which carries the discharge",
        },
        "flow__receiver_node": {
            "dtype": int,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "Node array of receivers (node that receives flow from current node)",
        },
        "topographic__elevation": {
            "dtype": float,
            "intent": "inout",
            "optional": False,
            "units": "m",
            "mapping": "node",
            "doc": "Land surface topographic elevation",
        },
        "topographic__steepest_slope": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The steepest *downhill* slope",
        },
        "bathymetric__depth": {
            "dtype": float,
            "intent": "in",
            "optional": False,
            "units": "-",
            "mapping": "node",
            "doc": "The water depth under the sea",
        },
    }

    def __init__(
        self,
        grid,
        repose_angle_cont=2.0,
        repose_angle_mar=2.0,
        porosity=0.0,
        max_erosion_rate=0.01,
        active_layer_rate=None,
        fields_to_track=None,
    ):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        repose_angle_cont : float or array-like (degree)
            The static angle of repose in the continental domain for one or
            multiple lithologies above which sediments move downslope.
        repose_angle_mar : float or array-like (degree)
            The static angle of repose in the continental domain for one or
            multiple lithologies above which sediments move downslope.
        porosity : float or array-like (-)
            The porosity of the sediments at the time of deposition for one or
            multiple lithologies. When computing the active layer, this porosity
            is used unless the field 'sediment__porosity' is being tracked in
            the stratigraphy.
        max_erosion_rate : float (m/time), optional
            The maximum erosion rate of the sediments. If None, all the sediments
            may be eroded in a single time step. The erosion rate defines the
            thickness of the active layer if `active_layer_rate` is None.
        active_layer_rate : float or array-like (m/time), optional
            The rate of formation of the active layer, which is used to determine
            the composition of the transported sediments. By default, it is set
            by the maximum erosion rate.
        fields_to_track : str or array-like, optional
            The name of the fields at grid nodes to add to the StackedLayers at
            each iteration.
        """
        super().__init__(grid, 1, porosity, fields_to_track)

        # Parameters
        self._repose_angle_cont = np.deg2rad(convert_to_array(repose_angle_cont))
        self._repose_angle_mar = np.deg2rad(convert_to_array(repose_angle_mar))
        self._max_erosion_rate = (
            np.inf if max_erosion_rate is None else max_erosion_rate
        )
        self._active_layer_rate = (
            self._max_erosion_rate if active_layer_rate is None else active_layer_rate
        )

        # Physical fields
        self._bathymetry = grid.at_node["bathymetric__depth"]

        # Field for the steepest slope
        n_nodes = grid.number_of_nodes
        if isinstance(grid, RasterModelGrid):
            self._link_lengths = grid.length_of_d8
        else:
            self._link_lengths = grid.length_of_link
        self._node_order = np.zeros(n_nodes, dtype=int)
        self._flow_receivers = grid.at_node["flow__receiver_node"]
        self._link_to_receiver = grid.at_node["flow__link_to_receiver_node"]
        self._slope = grid.at_node["topographic__steepest_slope"]
        if self._flow_receivers.ndim == 1:
            self._flow_receivers = self._flow_receivers[:, np.newaxis]
            self._link_to_receiver = self._link_to_receiver[:, np.newaxis]
            self._slope = self._slope[:, np.newaxis]
        self._steepest_receivers = np.zeros((n_nodes, 1), dtype=int)
        self._steepest_slope = np.zeros((n_nodes, 1))
        self._steepest_link = np.zeros((n_nodes, 1), dtype=int)

        # Fields for sediment fluxes
        n_sediments = self._stratigraphy.number_of_classes
        self._repose_angle = np.zeros((n_nodes, n_sediments))
        self._repose_slope = np.zeros((n_nodes, 1))
        self._slope_difference = np.zeros((n_nodes, 1))
        self._max_sediment_outflux = self._max_sediment_outflux[:, np.newaxis]
        self._active_layer = np.zeros((n_nodes, n_sediments))
        self._active_layer_thickness = np.zeros((n_nodes, 1))
        self._active_layer_composition = np.zeros((n_nodes, n_sediments))

    def _calculate_sediment_repose_angle(self):
        """
        Calculates the repose angle of the sediments over the continental and
        marine domains.
        """
        self._repose_angle[self._bathymetry == 0.0] = self._repose_angle_cont
        self._repose_angle[self._bathymetry > 0.0] = self._repose_angle_mar

    def _calculate_active_layer(self, max_thickness_sed):
        """
        Calculates the active layer based on the sediments.
        """
        porosity = (
            "sediment__porosity"
            if "sediment__porosity" in self._stratigraphy._attrs
            else self._porosity
        )
        self._active_layer[self._grid.core_nodes] = self._stratigraphy.get_active_layer(
            max_thickness_sed, porosity
        )

    def _calculate_active_layer_composition(self, dt):
        """
        Calculates the composition of the active layer based on the sediments
        and the bedrock.
        """
        self._calculate_active_layer(self._active_layer_rate * dt)

        self._active_layer_thickness[:] = np.sum(
            self._active_layer, axis=1, keepdims=True
        )
        self._active_layer_composition[:] = 0.0
        np.divide(
            self._active_layer,
            self._active_layer_thickness,
            out=self._active_layer_composition,
            where=self._active_layer_thickness > 0.0,
        )

    def _calculate_sediment_outflux(self, dt):
        """
        Calculates the sediment outflux for multiple lithologies.
        """
        self._calculate_sediment_repose_angle()
        self._calculate_active_layer_composition(dt)

        self._steepest_receivers[:] = np.argmax(self._slope, axis=1, keepdims=True)
        self._steepest_slope[:] = np.take_along_axis(
            self._slope, self._steepest_receivers, axis=1
        )
        self._steepest_link[:] = np.take_along_axis(
            self._link_to_receiver, self._steepest_receivers, axis=1
        )
        self._repose_slope[:] = np.tan(
            np.sum(
                self._repose_angle * self._active_layer_composition,
                axis=1,
                keepdims=True,
            )
        )
        self._slope_difference[:] = self._steepest_slope - self._repose_slope
        self._slope_difference[self._slope_difference < 0.0] = 0.0

        self._sediment_outflux[:, 0] = (
            self._active_layer_composition
            * self._slope_difference
            * self._link_lengths[self._steepest_link]
            * self._grid.cell_area_at_node[:, np.newaxis]
            / dt
        )

    def _threshold_sediment_outflux(self, dt):
        """
        Thresholds the sediment outflux to not go beyond the amount of sediments
        available for each lithology.
        """
        cell_area = self._grid.cell_area_at_node[:, np.newaxis]

        if self._max_erosion_rate != self._active_layer_rate:
            self._calculate_active_layer(self._max_erosion_rate * dt, 0.0)
        self._max_sediment_outflux[:, 0] = cell_area * self._active_layer / dt

        self._sediment_outflux[self._sediment_outflux > self._max_sediment_outflux] = (
            self._max_sediment_outflux[
                self._sediment_outflux > self._max_sediment_outflux
            ]
        )

    def run_one_step(self, dt, update_compatible=False, update=False):
        """Run the landslider for one timestep, dt.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        update_compatible : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer at the top of the stack only
            if the new layer is compatible with the existing layer.
        update : bool, optional
            If False, create a new layer and deposit in that layer; otherwise,
            deposition occurs in the existing layer.
        """
        self._calculate_sediment_outflux(dt)
        self._threshold_sediment_outflux(dt)

        self._node_order[:] = np.argsort(self._topography)
        self._sediment_influx[:] = 0.0
        calculate_sediment_influx(
            self._node_order,
            np.take_along_axis(self._flow_receivers, self._steepest_receivers, axis=1)[
                :, 0
            ],
            self._link_lengths[self._steepest_link][:, 0],
            self._grid.cell_area_at_node,
            self._steepest_slope[:, 0],
            self._repose_angle,
            self._sediment_influx,
            self._sediment_outflux[:, 0],
            dt,
        )

        self._apply_fluxes(dt, update_compatible, update)



from .water_driven_diffuser import WaterDrivenDiffuser
from .water_driven_router import WaterDrivenRouter
from .flux_driven_router import FluxDrivenRouter

__all__ = ["WaterDrivenDiffuser", "WaterDrivenRouter", "FluxDrivenRouter"]
